#
# Fetch utilities
#
# Note some functions are prefixed by wget, implying we use wget while, in fact
# we use curl (when available)... Perhaps we should add curl-native as a dependency
# for native-toolset so that we will no more have to revert to wget?...
#

downloadDir=$DOWNLOAD_LOC
srcUrl="$SRC_LOC"
srcSha256=$SRC_SHA256
srcMd5=$SRC_MD5
message "downloadDir='$downloadDir'"
message "srcUrl='$srcUrl'"

# Detect possible parameters and save them (this is needed for SCMS as the scheme name 
# the scms and the protocol option gives the protocol tu use
srcParams="${srcUrl#*;}"
srcUrl=${srcUrl%%;*}

# Clear all known parameters
for param in \
	apply striplevel patchdir mindate maxdate minrev maxrev rev notrev module \
	unpack destsuffix subdir localdir subpath name downloadfilename protocol
do
	eval unset src${param^}
done
if [ "$srcUrl" != "$srcParams" ]; then
	srcParams="${srcParams//;/ }"
	for param in $srcParams; do
		eval src${param^}
		message "src${param^}"
	done
fi
if [ "${srcUnpack:-unset}" = "unset" ]; then
	if [ "$DO_NOT_EXTRACT_SOURCE" = "true" ]; then
		srcUnpack=false
	else
		srcUnpack=true
	fi
	message "srcUnpack=$srcUnpack"
fi

# For relative URLs, add the workspace location in prefix
case $srcUrl in
	*:*)
		;;
	/*)
		;;
	empty)
		;;	
	*)
		srcUrl=$WORKSPACE_LOC/$srcUrl
		;;
esac

if [ -n "$srcDownloadfilename" ]; then
	srcFilename="$srcDownloadfilename"
else
	srcFilename=${srcUrl##*/}
fi
require_script_once "scms_utils.sh"
srcFile="$downloadDir/$srcFilename"

#	
# WARNING:
#	Fetched source is often shared between workspaces of the same 
# 	or different users, possibly even on different machines.
#	We must then lock Fetch and Import on the same source
#	file to avoid conflicts (simultaneously writing the same
#	files or writing files that are being read...) so we
#	hold an lock on the source file marker during these tasks
#
if [ -z "$BUILDSTEP_LOCK_FILE" ]; then
	if [ "$current_buildstep" = "fetch" ]; then
		BUILDSTEP_LOCK_FILE="$srcFile.done"
	elif [ "$current_buildstep" = "import" ]; then
		BUILDSTEP_LOCK_FILE="+$srcFile.done"
	fi
fi

require_script_once mirrors_utils.sh

gen_sha256() {
	local src="$1"
	local computed_sha256=$(sha256sum -b "$src")
	echo ${computed_sha256%% *}
}

gen_md5() {
	local src="$1"
	local computed_md5=$(md5sum -b "$src")
	echo ${computed_md5%% *}
}

gen_sha256_dir() {
	local src="$1"
	local sumfile=/tmp/sha256file.$$
	local fil
	sha256sum -b $(find $src -type f | sort | uniq) > $sumfile
	local computed_sha256=$(sha256sum -b "$sumfile")
	echo ${computed_sha256%% *}
}

gen_md5_dir() {
	local src="$1"
	local sumfile=/tmp/sha256file.$$
	local fil
	md5sum -b $(find $src -type f | sort | uniq) > $sumfile
	local computed_md5=$(md5sum -b "$sumfile")
	echo ${computed_md5%% *}
}
	

#
# Getting files using wget
#
wget_check() {
	local srcUrl=$1
	
	# wget seems to fail when there is a redirect to AWS...
	# We may try 
	#	curl --head --insecure --location --silent
	# where
	#	--head     (-I) only get the headers, testing for presence of the file
	#	--insecure (-k) will not verify certificates
	#	--location (-L) will obey 3xx answers for relocations
	#	--silent   (-s) eliminates output
	# but this fails too, as AWS fails the HEAD request (403 - Forbidden), so
	# we do a GET, and avoid transfering effectively the file by arranging to
	# get a write error (but only if we write effectively something, meaning 
	# writing inside a non-existant subdirectory does not work as we get the
	# error when creating th efile, not when writing) and check the  error
	# code returned bu curl is 23 meaning "write error" :-) (otherwise we
	# will probably get a 0 return status, or 22 if specifying --fail but
	# curl documentation says --fail is not reliable, so better avoid it...)
	#
	# To allow curl to start writing an get a write error soon so that it does
	# not transfer the whole file, we use "head -c 1" to close the pipeline
	# after 1 byte was transfered; closing the pipeline before curl starts
	# to write will not be detected the same by curl and will generate an
	# error status or 0 ;-(
	#
	
	# First test if curl is there
	if curl --version > /dev/null 2>&1; then
		curl --get --insecure --location --silent "$srcUrl" | head -c 1 > /dev/null 2>&1
		local status=${PIPESTATUS[0]} # note $? == ${PIPESTATUS[1]} is the status of head
		if [ $status -eq 23 ]; then
			# curl was able to fetch and start saving 1 byte, then get a write error,
			# meaning the target resource effectively exists
			true
		else
			# curl had nothing to write... The only case where this is not a fetch
			# error is if the resource exists but is empty, very unlikely here :-)
			false
		fi
	else
		# revert to wget, should only occur at the very beginning, before curl-native
		# can be installed
		wget --spider --no-check-certificate "$srcUrl" > /dev/null 2>&1
	fi
}

wget_download() {
	local srcFile=$1
	local destDir=$(dirname $srcFile)
	local srcUrl=$2
	
	mkdir -p "$destDir"
	# First test if curl is there
	if curl --version > /dev/null 2>&1; then
		curl --get --insecure --location --silent --output "$srcFile" "$srcUrl" 
	else
		# Don't check HTTPS certificate: we will check the downloaded file MD5/SHA256 signatures
		wget --no-check-certificate -nv --output-document="$srcFile" "$srcUrl"
	fi
}

wget_fetch() {
	local srcFile=$1
	local destDir=$(dirname $srcFile)
	local srcUrl=$2
	
	# Unset LD_LIBRATY_PATH to avoid error with sysroot OPENSSL; should no more be needed
	#unset LD_LIBRARY_PATH
	
	# First silently download from mirrors, as some official websites may redirect
	# non-current releases to an HTML page asking you to manually download the current
	# release or search for the desired one... (that's the case for the ICU website:
	# http://download.icu-project.org/files/icu4c/62.1/icu4c-62_1-src.tgz silently
	# redirects to http://site.icu-project.org/download, the download page of the ICU 
	if wget_from_mirror "$srcFile" "$srcUrl" false true $quiet; then
		return 0
	fi
	
	# Check original URL
	# We accept bad certificates for https URIs,
	# as anyway we'll check the fetched file MD5 and SHA256
	log "Check URL: '$srcUrl'"
	if wget_check $srcUrl; then
		wget_download "$srcFile" "$srcUrl"
	else
		warning "Can't get file from '$srcUrl', looking in mirrors"
		wget_from_mirror "$srcFile" "$srcUrl" # false false false
	fi
}

# Test if curl is not there
if ! curl --version > /dev/null 2>&1; then
	if ! wget --version > /dev/null 2>&1; then
		fatal 127 "System Workbench for Linux needs at least curl or wget to be installed; please install curl"
	else
		warning "System Workbench for Linux works better with curl than wget; please install curl if possible"
	fi
fi


check_downloaded_file() {
	local srcFile="$1"
	local cksumerror=${2:-false}
	local computed_sha256 computed_md5 file_sha256 done_sha256 done_md5
	
	if [ -z "$srcUrl" ]; then return 1; fi
	
	if [ -n "$scms_tool" ]; then
		if ${scms_tool}_check "$srcUrl" "$downloadDir" "$srcFilename"; then
			return 0
		else
			return 1
		fi
	fi
		
	case $srcUrl in
		http://* | https://* | ftp://* | file://*)
			if [ ! -r "$srcFile" ]; then
				return 1
			fi
			computed_sha256=$(gen_sha256 "$srcFile")
			computed_md5=$(gen_md5 "$srcFile")
			local found=true
			if [ -n "$srcSha256" ]; then
				if [ "$srcSha256" != "$computed_sha256" ]; then
					if $cksumerror; then
						error 1 "Sha256 checksum for '$srcFile' is wrong: expected '$srcSha256', found '$computed_sha256'"
					else
						warning "Found '$srcFile' but sha256 checksum is wrong: expected '$srcSha256', found '$computed_sha256'"
					fi
					found=false
				fi
			else
				warning "No sha256 checksum provided for '$srcUrl'; think at adding one for security"
				warning "Sha256 checksum for '$srcUrl' should be '$computed_sha256'"
			fi
			if [ -n "$srcMd5" ]; then
				if [ "$srcMd5" != "$computed_md5" ]; then
					if $cksumerror; then
						error 1 "Md5 checksum for '$srcFile' is wrong: expected '$srcMd5', found '$computed_md5'"
					else
						warning "Found '$srcFile' but md5 checksum is wrong: expected '$srcMd5', found '$computed_md5'"
					fi
					found=false
				fi
			else
				warning "No md5 checksum provided for '$srcUrl'; think at adding one for security"
				warning "Md5 checksum for '$srcUrl' should be '$computed_md5'"
			fi
			if $found; then
				(echo "MD5=$computed_md5"; echo "SHA256=$computed_sha256") > $srcFile.done
				(echo "MD5=$computed_md5"; echo "SHA256=$computed_sha256") > $(recipe_task_state_file).outstamp
				return 0
			fi
			;;
			
		*:*)# Non supported protocol
			;;
			
		*)
			if [ -d "$srcUrl" ]; then
				# Local directory, typically an Eclipse project in the workspace
				computed_sha256=$(gen_sha256_dir "$srcUrl")
				computed_md5=$(gen_md5_dir "$srcUrl")
				echo $computed_sha256 > "$srcFile.src-sha256"
				echo $computed_md5 > "$srcFile.src-md5"
				local found=false
				for fil in "$srcFile" "$srcFile.tar"* "${srcFile%.tar*}-$PACKAGE_VERSION.tar"*; do
					if [ -r "$fil" ]; then
						file_sha256=$(cat "$fil.src-sha256" 2> /dev/null)
						file_md5=$(cat "$fil.src-md5" 2> /dev/null)
						if [ "$file_sha256" = "$computed_sha256" -a "$file_md5" = "$computed_md5" ]; then
							computed_sha256=$(gen_sha256 "$fil")
							computed_md5=$(gen_md5 "$fil")
							(echo "MD5=$computed_md5"; echo "SHA256=$computed_sha256") > $fil.done
							(echo "MD5=$computed_md5"; echo "SHA256=$computed_sha256") > $(recipe_task_state_file).outstamp
							found=true
						else
							rm -f "$fil" "$fil.src-sha256" "$fil.src-md5"
						fi
					fi
				done
				if $found; then
					return 0
				fi
			else
				# Local tarball?
				local found
				for fil in "$srcFile" "$srcFile.tar"* "${srcFile%.tar*}-$PACKAGE_VERSION.tar"*; do
					if [ -r "$fil" ]; then
						computed_sha256=$(gen_sha256 "$fil")
						computed_md5=$(gen_md5 "$fil")
						local remove=false
						if [ -n "$srcSha256" -a "$srcSha256" != "$computed_sha256" ]; then
							warning "Found $(basename $fil), but with incorrect SHA256: $computed_sha256 instead of $srcSha256"
							remove=true
						fi
						if [ -n "$srcMd5" -a "$srcMd5" != "$computed_md5" ]; then
							warning "Found $(basename $fil), but with incorrect MD5: $computed_md5 instead of $srcMd5"
							remove=true
						fi
						if $remove; then
							# Ensure the marker file is empty, but don't delete it, it is used for locking fetch/import
							true > $fil.done
							continue
						fi
						found=$fil
					fi
				done
				if [ -n "$found" ]; then
					if [ -n "$srcSha256" ]; then
						warning "No sha256 checksum provided for '$found'; think at adding one for security"
						warning "Sha256 checksum for '$srcUrl' should be '$computed_sha256'"
					fi
					if [ -n "$srcMd5" ]; then
						warning "No md5 checksum provided for '$found'; think at adding one for security"
						warning "Md5 checksum for '$srcUrl' should be '$computed_md5'"
					fi
					(echo "MD5=$computed_md5"; echo "SHA256=$computed_sha256") > $found.done
					(echo "MD5=$computed_md5"; echo "SHA256=$computed_sha256") > $(recipe_task_state_file).outstamp
					return 0
				fi
			fi
			;;
	esac
	
	# Ensure the marker file is empty, but don't delete it, it is used for locking fetch/import
	true > $srcFile.done 
	rm -f $(recipe_task_state_file).outstamp
	remove_task_output_state
	return 1
}

base_do_fetch() {
	if check_downloaded_file "$srcFile"; then
		log "'$srcFilename' has been found in '$downloadDir'"
	elif [ -n "$scms_tool" ]; then
		if ${scms_tool}_fetch "$srcUrl" "$downloadDir" "$srcFilename"; then
			true
		else
			exit 1
		fi
	else
		case $srcUrl in
			http://* | https://* | ftp://*)
				log "Downloading $srcUrl ... into $downloadDir"
				rm -f "$srcFile" # Just in case the file exist but is not readable
				
				wget_fetch $srcFile $srcUrl
				check_downloaded_file "$srcFile" true
				;;
			file://*)
				rm -f "$srcFile" # Just in case the file exist but is not readable
				srcPath=${srcUrl#file://}
				case $srcPath in
					/*)
						log "Searching $srcPath on local machine"
						srcDir="$(dirname $srcPath)"
						;;
					*)
						log "Searching $srcPath in recipe resources"
						srcDir="$recipe_metadatadir/resources/$(dirname $srcPath)"
						;;
				esac
				srcPath=$(basename $srcPath)
				srcPath=$(machine_specific_file $srcDir $srcPath)
				if [ ! -r "$srcPath" ]; then
					warning "Resource file '$srcUrl' not found"
				fi
				if ! [ "$srcPath" -ef "$srcFile" ]; then
					cp "$srcPath" "$srcFile"
				fi
				log "Found '$srcUrl' in '$srcPath'"
				computed_sha256="$(gen_sha256 "$srcFile")"
				computed_md5="$(gen_md5 "$srcFile")"
				(echo "MD5=$computed_md5"; echo "SHA256=$computed_sha256") > $srcFile.done
				(echo "MD5=$computed_md5"; echo "SHA256=$computed_sha256") > $(recipe_task_state_file).outstamp
				;;
			*:*)
				# Ensure the marker file is empty, but don't delete it, it is used for locking fetch/import
				true > $srcFile.done 
				remove_task_output_state
				fatal 1 "Unsupported scheme: '${srcUrl%%:*}:'"
				;;
			*)
				if [ "$srcUrl" = "" ]; then
					fatal 1 "No source URL to fetch from..."
				fi
				if [ "$srcUrl" = "empty" ]; then
					srcDir="/tmp/fetch.$$"
					srcFilename="$srcUrl"
					srcUrl="$srcDir/$srcFilename"
					mkdir -p "$srcUrl"
				fi
				if [ -d "$srcUrl" ]; then
					# Local directory, typically an Eclipse project in the workspace
					if [ "$PACKAGE_VERSION" != "none" ]; then
						srcFile=${srcFile%-$PACKAGE_VERSION}-$PACKAGE_VERSION
					fi
					srcDir=${srcUrl%/*}
					rm -f "$srcFile.tar".*
					# Note that we do not pack the _build directory, that contain build history and will be
					# preserved when re-importing in the same workspace but is irrelevant in another workspace
					log "Packing '$srcUrl' in '$srcFile.tar.xz'"
					log_exec tar cJf "$srcFile.tar.xz" -C "$srcDir" --exclude=_build "$srcFilename"
					# Note that src-sha256 and src-md5 were computed before adding the version 
					mv "$(dirname $srcFile)/$srcFilename.src-sha256" "$srcFile.tar.xz.src-sha256"
					mv "$(dirname $srcFile)/$srcFilename.src-md5" "$srcFile.tar.xz.src-md5"
					computed_sha256="$(gen_sha256 "$srcFile.tar.xz")"
					computed_md5="$(gen_md5 "$srcFile.tar.xz")"
					(echo "MD5=$computed_md5"; echo "SHA256=$computed_sha256") > $srcFile.tar.xz.done
					(echo "MD5=$computed_md5"; echo "SHA256=$computed_sha256") > $(recipe_task_state_file).outstamp
				elif [ -r "$srcUrl" ]; then
					# Local tarball
					log "Copying '$srcUrl' in '$downloadDir'"
					log_exec cp "$srcUrl" "$downloadDir"
					computed_sha256="$(gen_sha256 "$srcFile")"
					computed_md5="$(gen_md5 "$srcFile")"
					(echo "MD5=$computed_md5"; echo "SHA256=$computed_sha256") > $srcFile.done
					(echo "MD5=$computed_md5"; echo "SHA256=$computed_sha256") > $(recipe_task_state_file).outstamp
				else
					fatal 2 "Can't find '$srcUrl'"
				fi
				;;
		esac
	fi
}
