#!/bin/bash
#
# Common Linux kernel and modules utilities
#

require_script_once "configure_utils.sh"

#	
# WARNING:
#	Simultaneously building modules and the DTB (or in fact running 
# 	two different make command simultaneously) may fail as some
#	files may be regenerated in parallel... We don't have to
#	protect the compile-image task as it is a dependency of
#	compile-modules and compile-dtb, but we must ensure these
#	two tasks will not run concurrently
#
if [ -z "$BUILDSTEP_LOCK_FILE" ]; then
	if [ "$current_buildstep" = "compile-modules" ]; then
		BUILDSTEP_LOCK_FILE="$current_task_statefile"
	elif [ "$current_buildstep" = "compile-device-tree" ]; then
		BUILDSTEP_LOCK_FILE="$(recipe_task_state_file compile-modules)"
	fi
fi

# Default variables
: ${SCRIPT_RESOURCES_FOLDER:=.}
CONFIGS_LOC=$SCRIPT_RESOURCES_FOLDER/configs
CROSS_COMPILE=${NATIVE_SYSROOT_LOC%/}/${compiler_loc%/}/usr/bin/${compiler_prefix}-
bootDir=$installDir/boot
dtsDir=arch/${ARCH}/boot/dts/

# Get the list of dts/dtsi from the project location
list_dts=(`find ./ -maxdepth 1 -name "*.dts" | sed 's/^\.\///g'`)
list_dtfiles=(`find ./ -maxdepth 1 -type f -regex ".*/.*\.\(dts\|dtsi\)"`)

do_move_dts() {
	# Copy dt files from source root to arch directory if exists
	if [ ${#list_dtfiles[@]} -gt 0 ]; then
		for dt_file in ${list_dtfiles[@]} ; do
			log_exec cp ${dt_file} ${dtsDir}
		done
	fi
}

do_make_config() {
	# Rebuild configuration, usually called for a default config or with oldconfig/silentoldconfig
	log_exec cd $srcDir
	local what="${1:oldconfig}"
    local bDir=$(cd $buildDir; /bin/pwd) sDir=$(/bin/pwd)
    if [ "$bDir" != "$sDir" ]; then
		log_exec make ARCH=${ARCH} O=$buildDir $what
	else
		log_exec make ARCH=${ARCH} $what
	fi
}

do_default_config() {
	# Apply default configuration
	do_make_config $defconfig
}

do_apply_config_fragments() {
	# Edit .config with selected kernel configuration fragments
	local config_fragment config_fragments
	log_exec cd $srcDir
    for config_fragment in ./fragments/*.config "$@"; do
	    if [ -f "$config_fragment" ]; then
	    	config_fragments="$config_fragments $config_fragment"
	    fi
	done
	    
	if [ "$buildDir" -ef "$srcDir" ]; then
    	log_exec ./scripts/kconfig/merge_config.sh .config $config_fragments
    else
    	log_exec ./scripts/kconfig/merge_config.sh -O $buildDir $buildDir/.config $config_fragments
    fi | grep -v '^Value ' | grep -v ' value: ' | uniq
}

get_kernel_version() {
	local ksrc="${1:-.}"
	if [ -d "$ksrc" ]; then
		if [ -r "$ksrc/include/generated/utsrelease.h" ]; then
			sed -n -e 's/^.*UTS_RELEASE "\([^"][^"]*\)".*$/\1/p' < "$ksrc/include/generated/utsrelease.h"
		fi
	fi
}

kernel_do_make() {
	do_make ARCH=${ARCH} CROSS_COMPILE=${CROSS_COMPILE} "$@"
}

do_make_image() {
	kernel_do_make ${image_format} "$@"
}

do_make_dts() {
	# Build the device tree from kernel recipse
	if [ ! -z ${device_name} ]; then
		kernel_do_make ${device_name}.dtb "$@"
	fi
	
	# Build the device tree from kernel resources
	if [ ${#list_dts[@]} -gt 0 ]; then
		for dts_file in ${list_dts[@]} ; do
			dtb_file=${dts_file/.dts/.dtb}
			kernel_do_make ${dtb_file} "$@"
		done
	fi
}

do_make_modules() {
	kernel_do_make modules "$@"
}

do_make_modules_install() {
	kernel_do_make INSTALL_MOD_PATH=${installDir} modules_install "$@"
}

do_install_kernel() {
	cd $buildDir
	
	log "Copy Linux Kernel image"
	log_exec install -d $bootDir
	log_exec install vmlinux $bootDir
	log_exec install System.map $bootDir
	log_exec install arch/${ARCH}/boot/Image $bootDir
	if [ "${image_format}" != "Image" ]; then 
		log_exec install arch/${ARCH}/boot/${image_format} $bootDir
	fi
}

do_install_dts() {	
	log "Copy Device Tree"
	log_exec install -d $bootDir
	
	# Install every dtb found in dtb build directory
	list_dtb=(`find $buildDir/arch/${ARCH}/boot/dts -name "*.dtb"`)
	if [ ${#list_dtb[@]} -gt 0 ]; then
		for dtb_file in ${list_dtb[@]} ; do
			log_exec install ${dtb_file} $bootDir
		done
	fi
}

do_install_modules() {
	log "Copy Modules files"
	do_make_modules_install "$@"
}
