######################################################################################################################
#                                                                                                                    #
#                                Generic script routines usable in generation scripts                                #
#                                                                                                                    #
######################################################################################################################

# Platform-specific variables
platform_loc=$PLATFORM_LOC
platformDir=$platform_loc
platform_id=$(basename $platform_loc)
platform_name=${platform_id%%_*}
platform_version=${platform_id#*_}

# Follow symbolic links to the download location to avoid false change indications
if [ -n "$DOWNLOAD_LOC" ]; then
	DOWNLOAD_LOC=$(absolute_path $DOWNLOAD_LOC)
fi

# Recipe specific variables
export recipe_id=$RECIPE_ID
export recipe_name=$RECIPE_NAME
export recipe_base_name=$recipe_name
export recipe_version=$RECIPE_VERSION
export recipe_revision=$RECIPE_REV

# Backward compatibility settings
: ${recipe_id:=$PACKAGE_ID}
: ${recipe_name:=$PACKAGE_NAME}
: ${recipe_version:=$PACKAGE_VERSION}
: ${recipe_revision:=$PACKAGE_REV}
recipe_spec=$recipe_name
if [ -n "$recipe_version" ]; then
	recipe_spec="${recipe_spec}_${recipe_version}"
fi

if [ -n "$BUILDSTEP_SCRIPT_PATH" ]; then
	current_buildstep_script="$BUILDSTEP_SCRIPT_PATH"
else
	current_buildstep_script=$(readlink -m "$0")
fi
buildstep_script=$(basename "$current_buildstep_script" .sh)

tmpdir="/tmp/tmpdir"
if [ -n "$platform_id" ]; then 
	tmpdir="${tmpdir}.${platform_id}"
fi
if [ -n "$recipe_id" ]; then 
	tmpdir="${tmpdir}.${recipe_id}"
fi
if [ -n "$buildstep_script" ]; then 
	tmpdir="${tmpdir}.${buildstep_script}"
fi
tmpdir="${tmpdir}.$$"

rm -rf $tmpdir
mkdir -p $tmpdir
# We must cleanup our temporary directory when exiting
record_cleanup rm -rf $tmpdir


######################################################################################################################
#                                                                                                                    #
#                                                     Global setup                                                   #
#                                                                                                                    #
######################################################################################################################

#
# Ensure we get shared library from the native sysroot before the installed ones or sysroot-installed
# native tools may not work as expected...
#
# THIS IS NO MORE NEEDED as we set RPATH in all -native executables
#for dir in /usr/lib /lib $(cat $NATIVE_SYSROOT_LOC/etc/ld.so.conf 2>/dev/null); do
#	dir=${dir#$NATIVE_SYSROOT_LOC}
#	if [ -d "$NATIVE_SYSROOT_LOC$dir" ]; then
#		LD_LIBRARY_PATH="$NATIVE_SYSROOT_LOC$dir:$LD_LIBRARY_PATH"
#	fi
#done
# On the opposite ensure LD_LIBRARY_PATH do not reference the native sysroot; this is needed for now as
# the IDE sets it...
suppress_tree_from_path() { # Also suppress empty entries (meaning current directory)
	local path="$1"
	local what="${2:-$NATIVE_SYSROOT_LOC}"
	local dir sep=''
	for dir in ${path//:/ }; do
		if [ -z "$dir" -o "$dir" = "null" ]; then continue; fi
		if [[ " $dir/" =~ " ${what%/}/" ]]; then
			continue
		fi
		echo -n "$sep$dir"
		sep=':'
	done
}

#
# When running under fakeroot, we have LD_PRELOAD set, but we need the LD_LIBRARY_PATH setup
# to be preserved, or these libraries can't be loaded. The problem is that some recipe build
# systems (like Python) set LD_LIBRARY_PATH instead of augmenting it, so fakeroot fails!...
# To avoid that we resolve the LD_PRELOAD libraries now, before LD_LIBRARY_PATH was cleaned.
#
resolve_libraries() {
	local path lib preload
	for lib; do
		path=$(ldd /bin/true | sed -n -e "s/^[[:space:]]*$lib[[:space:]]*=>[[:space:]]*\([^[:space:]]*\).*$/\1/p")
		preload="$preload $path"
	done
	echo ${preload}
}

ORIGINAL_LD_LIBRARY_PATH="$LD_LIBRARY_PATH";                 export ORIGINAL_LD_LIBRARY_PATH
LD_LIBRARY_PATH=$(suppress_tree_from_path $LD_LIBRARY_PATH); export LD_LIBRARY_PATH
LD_PRELOAD=$(resolve_libraries "${LD_PRELOAD//:/ }");        export LD_PRELOAD
# echo ORIGINAL_LD_LIBRARY_PATH="$ORIGINAL_LD_LIBRARY_PATH" >&2
# echo LD_LIBRARY_PATH="$LD_LIBRARY_PATH" >&2

# 
# We don't need to augment the LD_LIBRARY_PATH, however we have to install some native tools in a "fake" 
# or "simplified" version to allow building other tools and then build the "full" version of the tool.
# That is for example the case with texinfo-native, that need the autotools that require texinfo, so we
# provide them with a fake texinfo-native, named texinfo-dummy-native, that is installed in the weakbindir
#
# We must then place this weak sysroot in our PATH, just after the native sysroot
#
# Note that dummy recipes will populate the weak sysroot instead of the native one; they shoudl only
# install binary programs; if such a recipe install a shared library, it will not be accessible by
# standard programs, nor even probably by dummy programs...
#
WEAK_SYSROOT_LOC="${NATIVE_SYSROOT_LOC%/}"
WEAK_SYSROOT_LOC="${WEAK_SYSROOT_LOC%/native}/weak"
export WEAK_SYSROOT_LOC

# For Yocto compatibility we must have a few host tools found in a
# platform-local directory inserted in our PATH.
# Note however that, being at the end of the PATH, we will never use them,
# so this code will mainly check all needed host tools are present ;-)
HOSTTOOLS="[ ar as awk basename bash bzip2 cat chgrp chmod chown chrpath \
           cmp comm cp cpio cpp cut date dd diff diffstat dirname du echo \
           egrep env expand expr false fgrep file find flock g++ gawk gcc \
           getconf getopt git grep gunzip gzip head hostname id install ld \
           ldd ln ls make makeinfo md5sum mkdir mknod mktemp mv nm objcopy \
           objdump od patch perl pod2man pr printf pwd python2 python2.7 \
           python3 ranlib readelf readlink rm rmdir rpcgen sed sh sha256sum \
           sleep sort split stat strings strip tail tar tee test touch tr \
           true uname uniq wc wget which xargs"
HOSTTOOLS_DIR=
HOSTTOOLS_DIR="${NATIVE_SYSROOT_LOC%/}"
HOSTTOOLS_DIR="${HOSTTOOLS_DIR%/native}/host-tools"
mkdir -p "$HOSTTOOLS_DIR"
for hosttool in $HOSTTOOLS; do
	hosttool_bin=$(which $hosttool 2>/dev/null)
	if [ -z "$hosttool_bin" ]; then
		fatal 99 "Host tool '$hosttool' is not found in PATH... Please install it"
	fi
	ln -nfs $hosttool_bin $HOSTTOOLS_DIR/$hosttool
done
PATH="$PATH:$HOSTTOOLS_DIR"

expand_path() {
	local prefix=$1
	local path="$2"
	local what="${3:-$NATIVE_SYSROOT_LOC}"
	local where="${4:-$WEAK_SYSROOT_LOC}"
	local dir sep='' all_dirs=""
	what="${what%/}"
	where="${where%/}"
	for dir in ${path//:/ }; do
		dir="${dir%/}"
		# Eliminate empty entries, inexistant entries and duplicates
		if [ -z "$dir" -o "$dir" = "null" ]; then continue; fi
		if [[ " $all_dirs " =~ " $dir " ]]; then continue; fi
		all_dirs="$all_dirs $dir"

		if [[ " $dir/" =~ " $what/" ]]; then
			local ndir="$where${dir#$what}"
			if [ -d "$ndir" ]; then
				if [ -d "$dir" ]; then
					if $prefix; then
						echo -n "$sep${ndir}:${dir}"
					else
						echo -n "$sep${dir}:${ndir}"
					fi
				else
					echo -n "$sep${ndir}"
				fi
			elif [ -d "$dir" ]; then
				echo -n "$sep${dir}"
			else
				continue
			fi
		else
			echo -n "$sep${dir}"
		fi
		sep=':'
	done
}

prefix_path() {
	expand_path true "$@"
}

postfix_path() {
	expand_path false "$@"
}

export ORIGINAL_PATH="$PATH"
PATH="$(postfix_path $PATH $NATIVE_SYSROOT_LOC $WEAK_SYSROOT_LOC)"

#
# Search or copy files looking first in the native sysroot, then 
# in the weak sysroot
#
search_in_native_sysroot() {
	local what="$1"
	if [ -r "$NATIVE_SYSROOT_LOC/$what" ]; then
		echo "$NATIVE_SYSROOT_LOC/$what"
	elif [ -r "$WEAK_SYSROOT_LOC/$what" ]; then
		echo "$WEAK_SYSROOT_LOC/$what"
	fi
}

copy_from_native_sysroot() {
	local from=$(search_in_native_sysroot "$1")
	local to="$2"
	if [ -r "$from" ]; then
		log_exec cp "$from" "$to"
	fi
}

#
# Compute the optimal build parallelism (this should work on most POSIX systems)
#
if [ -z "$CORE_COUNT" ]; then 
	CORE_COUNT=$(nproc || sysctl -n hw.cpu || getconf _NPROCESSORS_ONLN)
fi 2> /dev/null 

#
# First determine the current buildstep and script name
#
get_current_buildstep() {
	case $buildstep_script in
		shell)          current_buildstep="development-shell"           ;;
		patch)          current_buildstep="apply-patches"               ;;
		make_all)       current_buildstep="compile"                     ;;
		make_install)   current_buildstep="install"                     ;;
		*-*)            current_buildstep="${buildstep_script#*-}"      ;;
		*)              current_buildstep="${buildstep_script//_/-}"    ;;
	esac
}

if [ -n "$BUILDSTEP_SCRIPT_PATH" ]; then
	current_buildstep_script="$BUILDSTEP_SCRIPT_PATH"
else
	current_buildstep_script=$(readlink -m "$0")
fi
buildstep_script=$(basename "$current_buildstep_script" .sh)

current_buildstep=$CURRENT_BUILDSTEP
if [ -z "$current_buildstep" ]; then
	current_buildstep=$(get_current_buildstep)
fi
current_task="$current_buildstep" # Task and Buildstep are synonyms

# Also set some buildstep-specific settings
case $buildstep in
	install)	sequential_build=true	;; # Always install sequentially
esac

#
# Transform a entity name in a shell variable name
#
# NOTE: This should always be synchronized with import_package.sh
#
shell_variable_name() {
	local name="$1"
	name="${name//+/P}"
	name="${name//\./_DOT_}"
	name="${name//[^a-zA-Z0-9_]/_}"
	echo "$name"
}

prepend_to_shell_variable() {
	local varname="$(shell_variable_name $1)"
	shift
	eval "$varname=\"$* ${!varname}\""
}

append_to_shell_variable() {
	local varname="$(shell_variable_name $1)"
	shift
	eval "$varname=\"${!varname} $*\""
}

#
# Test if a shell function exist
#
fn_exists() {
	declare -f "$1" > /dev/null 2>&1
}

simplify_function_alternates() {
	local alt alternates
	for alt in ${1//-/_}; do # Function names may be built from task names...
		alternates="$alternates ${alt//-/_}" # Function names may be built from task names...
	done
	echo "${alternates# }"
}

#
# Call a shell function if it exists
#
# The set of functions to be tried must be passed as ONE SINGLE argument
# and only the first to exist will be called
#
shell_function_executed=false
call_if_exists() {
	dbg scripts "Searching for first function in set '$1'" 
	local alternates=$(simplify_function_alternates "$1")
	shift
	
	local func status
	for func in $alternates; do
		if fn_exists "$func"; then
			shell_function_executed=true
			note "Executing function '$func'"
			eval "$func" "$@"
			status=$?
			dbg_note "Done executing function '$func': status=$status"
			return $status
		fi
	done
	return 0
}

#
# Call all shell functions from recipe classes that exists (each may have alternate names)
#
call_all_existing() {
	dbg scripts "Searching in all classes for function in set '$1'" #
	local alternates=$(simplify_function_alternates "$1")
	shift
	
	local class status
	for class in $recipe_classes; do
		call_if_exists "${class}_${alternates// / ${class}_}" "$@"
		status=$?
		if [ "$status" -ne 0 ]; then return $status; fi
	done
	call_if_exists "${alternates}" "$@"
}

#
# Call the first shell functions from recipe classes that exists (it may have alternate names)
#
call_first_existing() {
	dbg scripts "Searching for first class having function in set '$1'" 
	local alternates=$(simplify_function_alternates "$1")
	shift
	
	local class status executed=$shell_function_executed
	shell_function_executed=false
	for class in $recipe_classes; do
		call_if_exists "${class}_${alternates// / ${class}_}" "$@"
		status=$?
		if [ "$status" -ne 0 ]; then return $status; fi
		if $shell_function_executed; then
			return $status
		fi
	done
	shell_function_executed=$executed
	call_if_exists "${alternates}" "$@"
}

#
# Call all shell functions from recipe classes that exists
#
call_all_existing_reverse() {
	dbg scripts "Searching in all classes (reversed order) for function in set '$1'" 
	local alternates=$(simplify_function_alternates "$1")
	shift
	
	local class status
	call_if_exists "${alternates}" "$@"
	status=$?
	if [ "$status" -ne 0 ]; then return $status; fi
	for class in $recipe_classes_reversed; do
		call_if_exists "${class}_${alternates// / ${class}_}" "$@"
		status=$?
		if [ "$status" -ne 0 ]; then return $status; fi
	done
	return 0
}

#
# Set a variable to a default value if it's not set
#
default_value_if_not_set() {
	local varname="$1"
	shift
	local defaultval="$*"
	if [ ! "${!varname+set}" = "set" ]; then
		# Variable $varname is not set, even to a null value
		eval $varname="'${defaultval}'"
	fi
}

#
# Inherit an unset variable from some other variable, if this one is set,
# or set it to a default value
#
default_value_if_not_set_or_inherited() {
	local varname="$1"
	local inheritvar="$2"
	shift 2
	local defaultval="$*"
	if [ ! "${!varname+set}" = "set" ]; then
		# Variable $varname is not set, even to a null value
		# First try to inherit from $inheritvar, if this one is set (maybe to null)
		if [ "${!inheritvar+set}" = "set" ]; then
			eval $varname="'${!inheritvar}'"
		else
			eval $varname="'${defaultval}'"
		fi
	fi
}

#
# Bash only supports ${parameter@operator} starting with bash-4.4; we only use
# the @ operator to requote strings so we must emulate it when using an older bash...
#
bash_version=$(bash --version | sed -n -e 's/^.* version \([0-9]\+\)\.[0-9]\+.*$/\1/p')
bash_subvers=$(bash --version | sed -n -e 's/^.* version [0-9]\+\.\([0-9]\+\).*$/\1/p')
if [ $bash_version -ge 5 ] || [ $bash_version -eq 4 -a $bash_subvers -ge 4 ]; then
	compat_requote() {
		echo "${*@Q}"
	}
else
	compat_requote() {
		local arg sep=
		for arg in "$@"; do
			printf "$sep%s" "'${arg//\'/\'\\\'\'}'"
			sep=" "
		done
		echo
	}
fi

requote_args() {
	compat_requote "$@"
}

#
# Requote smartly a single argument, quoting only the value of valued arguments
# but fully quoting arguments that contain spaces or other shell special
# characters; also unwrap existing quotes around the argument
#
requote_arg() {
    local arg="$*" opt val
    if [ -z "$arg" ]; then return 0; fi
    # Don't protect ${} so that variable references are expandable.
    # Use printf here so that initial -e or -n gets out correctly.
    # Note the first case alternative will probably never occur...
    case $arg in
		*\ *=* | *\(*=* | *\)*=* | *\!*=* | *\'*=* | *\"*=* | *\#*=*) #' # Some IDEs are confused here
		    printf "%s\n" "$(compat_requote "$arg")"
		    ;;
		-?*=*\ * | -?*=*\(* | -?*=*\)* | -?*=*\!* | -?*=*\'* | -?*=*\"* | -?*=*\#*) #' # Some IDEs are confused here
		    opt=${arg%%=*}
		    val=${arg#*=}
		    printf "%s=%s\n" $opt "$(compat_requote "$val")"
		    ;;
		*\ * | *\(* | *\)* | *\!*i | *\'* | *\"* | *\#*) #' # Some IDEs are confused here
		    printf "%s\n" "$(compat_requote "$arg")"
		    ;;
		*)
		    printf "%s\n" $arg
		    ;;
    esac
}

# 
# Split words from the arguments and outputs them one per line
#
# Splitting a variable on spaces while having quotes protect spaces is a bit
# tricky, as bash do not look at quotes when splitting words in a for loop...
#
# Typical use is to place words in an array by
#		mapfile -t array < <(split_in_words "$args")
#		mapfile -t array < <(split_in_words "$(compat_requote "$@"))
#	Beware that <(...) is NOT a redirection fo the outer process input but
#	of the inner process output, and expands as the pathname of the file
#   written by the inner process, so the two, SEPARATED, <
# 	The '-t' is there to trim the newline at the end of each array element
#
#	One can also write
#		split_in_words $args > $tmpfile
#		mapfile -p array < $tmpfile
#
#	Beware also that the more appealing
# 		split_in_words "$args" | mapfile -t array
#	will not work, creating array in a subshell created to host the RHS of the pipe
#
split_in_words() {
	# This eval construct is there to do variable expansion BEFORE quote
	# analysis, which is otherwise done first (while word splitting is done 
	# last, but taking care of space escaping done during the quote analysis
	# phase.
	if [ -z "$*" ]; then return 0; fi
	local loop="for arg in $*; do requote_arg \$arg; done"
	eval $loop
}

#
# Parse arguments, splitting each parameter on spaces obeying quotes to wrap arguments
#
parse_args_and_call() {
	local cmd=$1 arg
	shift
	declare -a args=()
	# Be careful to protect variable references here, they may not yet be valid...
	mapfile -t args < <( for arg; do split_in_words "$arg"; done )
	$cmd "${args[@]}"
}

# Consider only the first argument as a composite that must be parsed
parse_first_arg_and_call() {
	local cmd=$1 arg
	shift
	declare -a args=()
	# Be careful to protect variable references here, they may not yet be valid...
	mapfile -t args < <( split_in_words "$1" )
	shift
	# Be careful also not to generate an extra empty arg
	if [ $# -gt 0 ]; then
		$cmd "${args[@]}" "$@"
	else
		$cmd "${args[@]}"
	fi
}

######################################################################################################################
#                                                                                                                    #
#                                                Shared file management                                              #
#                                                                                                                    #
######################################################################################################################

#
# Merge lists of lines into a shared file, taking care of locking the shared file to
# avoid race conditions if two tasks try to modify th esame shared file simultaneously
merge_into_shared_list_file() {
	local result="$1"
	touch "$result"

	(
		flock -x 0
		cat "$result" "$@" | sort | uniq > "$result.new"
		if cmp -s "$result" "$result.new"; then
			rm -f "$result.new"
		else
			rm -f "$result"
			mv "$result.new" "$result"
		fi
	) < "$result"
}

######################################################################################################################
#                                                                                                                    #
#                                               Wrapper-script generation                                            #
#                                                                                                                    #
######################################################################################################################

create_cmdline_wrapper() {
	#
	# Create a wrapper script where extra command line arguments are needed
	#
	# These are useful to work around relocation issues, by passing extra
	# command line arguments to a program.
	#
	# Usage: create_cmdline_wrapper [ BASEDIR ] FILENAME extra-options...
	#
	# All extra-options are passed to the command in front of its other arguments
	#
	# The BASEDIR parameter can be omitted as long as create_wrapper is called
	# after installing an executable in the install directory (that is during the
	# install task); that is to be compatible with the create_wrapper function
	# called by Yocto install recipes.
	#
	local BASEDIR
	if [ $# -lt 2 ]; then
		fatal 1 "create_cmdline_wrapper needs at least 2 parameters: [ BASEDIR ] FILENAME ( argument | VAR=value )..."
	fi
	
	if [ -d "$1" ]; then
		BASEDIR=${1%/}
		shift
	else
		BASEDIR=$installDir
	fi
	
	local cmd=$1
	shift

	cmd=${cmd#$BASEDIR/}
	if [ ! -r $BASEDIR/$cmd ]; then
		fatal 2 "Can't find '$cmd' in '$BASEDIR'"
	fi

	echo "Generating wrapper script for $cmd"

	mv $BASEDIR/$cmd $BASEDIR/$cmd.real
	local cmdname=`basename $cmd`
	local dirname=`dirname $cmd`
	local relpath=$(get_up_path $dirname)
	local -a extra_args
	local argno=0
	
	echo '#! /bin/bash'                                                                     >  $BASEDIR/$cmd
	echo 'realpath="$(readlink -fn $0)"'                                                    >> $BASEDIR/$cmd
	echo 'realdir="$(dirname $realpath)"'                                                   >> $BASEDIR/$cmd
	for arg; do
		extra_args[argno++]=${arg//${BASEDIR//\//\\/}/\$realdir$relpath}
	done
	if [  $argno -ne 0 ]; then
		echo 'exec -a $realdir/'$cmdname '$realdir/'$cmdname.real "${extra_args[@]}" '"$@"' >> $BASEDIR/$cmd
	else
		echo 'exec -a $realdir/'$cmdname '$realdir/'$cmdname.real '"$@"'                    >> $BASEDIR/$cmd
	fi
	chmod +x $BASEDIR/$cmd
}

create_wrapper() {
	#
	# Create a wrapper script where extra environment variables or arguments are needed
	#
	# These are useful to work around relocation issues, by setting environment
	# variables which point to paths in the filesystem.
	#
	# Usage: create_wrapper [ BASEDIR ] FILENAME ( argument | VAR=value )...
	#	Arguments are passed in front of $cmd argument list
	#	Variable assignments are passed as environment variables
	#
	# The BASEDIR parameter can be omitted as long as create_wrapper is called
	# after installing an executable in the install directory (that is during the
	# install task); that is to be compatible with the create_wrapper function
	# called by Yocto install recipes.
	#
	local BASEDIR
	if [ $# -lt 2 ]; then
		fatal 1 "create_wrapper needs at least 2 parameters: [ BASEDIR ] FILENAME ( argument | VAR=value )..."
	fi
	
	if [ -d "$1" ]; then
		BASEDIR=${1%/}
		shift
	else
		BASEDIR=$installDir
	fi
	
	local cmd=$1
	shift

	cmd=${cmd#$BASEDIR/}
	if [ ! -r $BASEDIR/$cmd ]; then
		fatal 2 "Can't find '$cmd' in '$BASEDIR'"
	fi
	
	echo "Generating wrapper script for $cmd"

	mv $BASEDIR/$cmd $BASEDIR/$cmd.real
	local cmdname=`basename $cmd`
	local dirname=`dirname $cmd`
	local relpath=$(get_up_path $dirname)
	local -a extra_args
	local argno=0

	echo '#! /bin/bash'                                                                     >  $BASEDIR/$cmd
	echo 'realpath="$(readlink -fn $0)"'                                                    >> $BASEDIR/$cmd
	echo 'realdir="$(dirname $realpath)"'                                                   >> $BASEDIR/$cmd

	for arg; do
		arg=${arg//${BASEDIR//\//\\/}/\$realdir$relpath}
		case $arg in
			-*)
				extra_args[argno++]="$arg"
				;;
			*=*)
				local val="${arg#*=}"
				local valprog="${val%% *}"
				local argweak="${arg//${NATIVE_SYSROOT_LOC//\//\\/}\//$WEAK_SYSROOT_LOC/}"
				local valweak="${argweak#*=}"
				local weakprog="${valweak%% *}"
				if [ "$valprog" != "$weakprog" ]; then
					echo "if [ ! -r '$valprog' ] \\"
					echo "&& [ -r '$weakprog' ]; then"
					echo "	export $argweak"
					echo "else"
					echo "	export $arg"
					echo "fi"
				else	
					echo export "$arg"
				fi                                                                          >> $BASEDIR/$cmd
				;;
			*)
				extra_args[argno++]="$arg"
				;;
		esac
	done
	if [  $argno -ne 0 ]; then
		echo 'exec -a $realdir/'$cmdname '$realdir/'$cmdname.real "${extra_args[@]}" '"$@"' >> $BASEDIR/$cmd
	else
		echo 'exec -a $realdir/'$cmdname '$realdir/'$cmdname.real '"$@"'                    >> $BASEDIR/$cmd
	fi
	chmod +x $BASEDIR/$cmd
}

######################################################################################################################
#                                                                                                                    #
#                                                   Tarball management                                               #
#                                                                                                                    #
######################################################################################################################

build_tarball() {
	if [ $# -le 2 ]; then 
		internal_error "build_tarball must be called with at least 3 parameters (tarball, basedir, files...)"
	fi
	local tarball="$1"
	local srcDir="$2"
	shift 2
	case $tarball in 
		*.tar.lzo | *.tzo)	zopt="--lzop"	;;
		*.tar.gz | *.tgz)	zopt="-z"    	;;
		*.tar.bz2 | *.tbz2)	zopt="-j"    	;;
		*.tar.bz | *.tbz)	zopt="-j"    	;;
		*.tar.xz | *.txz)  	zopt="-J"    	;;
		*.tar.lzma | *.tlz) zopt="--lzma" 	;;
		*)               	zopt=        	;;
	esac
	log_run_locked $tarball tar cf $tarball $zopt -C "$srcDir" "$@"
}

######################################################################################################################
#                                                                                                                    #
#                                                  Directory management                                              #
#                                                                                                                    #
######################################################################################################################

#
# Check if a directory is empty
#
is_directory_empty() { 
	local dir="${1:-.}"
	if find $dir -mindepth 1 2> /dev/null | read; then
		return 1
	fi
	return 0
}

is_directory_not_empty() { 
	local dir="${1:-.}"
	find $dir -mindepth 1 2> /dev/null | read
}

#
# Check if a directory tree contains files or only subdirectories
#
is_tree_empty() { # Ignore the .ac6_build marker directories and .fakeroot overlays
	local dir="${1:-.}" line
	if [ -n "$dir" -a -d "$dir" ]; then 
		if find $dir -name .ac6_build -prune -o -type f -print | grep -v '/.fakeroot$' | read line; then
			return 1
		fi 2> /dev/null
	fi
	return 0
}

is_tree_not_empty() { # Ignore the .ac6_build marker directories and .fakeroot overlays
	local dir="${1:-.}" line
	if [ -n "$dir" -a -d "$dir" ]; then 
		if find $dir -name .ac6_build -prune -o -type f -print | grep -v '/.fakeroot$' | read line; then
			return 0
		fi 2> /dev/null
	fi
	return 1
}

#
# Copy a directory tree, excluding files/directories passed as additional parameters
# When possible files are linked instead of being copied, to save space on the disk
# Destination directory is created if needed
#
# Syntax: copy_tree_excluding options source destination files_to_exlude...
# Options:
#	-v		Print the list of copied files on output
#	-d		Create all directories found in source, even empty directories
#	-l dir	Lock this directory when doing the shallow copy (default is source directory)
#
# NOTE: This does not copy any file whose name may start by one of the patterns_to_exclude
# or /.ac6_build/ and /.fakeroot
#
# WARNING: this does a shallow copy, not copying files but creating hard links; this
# has the undesirable effect of modifying the files inodes, so we must lock the 
# original source directory otherwise a real copy will choke on the file being modified
# while being read 
#
copy_tree_excluding() {
	local cpio_opts="-pdlmu"
	local find_opts="-depth -not -type d"
	local lock
	while true; do
		if [ "$1" = "-v" ]; then 
			cpio_opts="-pdlmv"
			shift
		elif [ "$1" = "-d" ]; then 
			find_opts="-depth"
			shift
		elif [ "$1" = "-l" ]; then 
			if [ $# -ge 2 ]; then
				lock="$2"
				shift 2
			else # Just ignore it
				shift
			fi
		elif [[ " $1" =~ " -l" ]]; then
			lock="${1#-l}"
			shift
		else
			break
		fi
	done
	local src="${1%/}"
	local dst="${2%/}"
	shift 2
	local exclude sedopts="-e 's|^./||'"
	for exclude in '/.ac6_build/' '/.fakeroot' "$@"; do # Never shallow copy .ac6_build or .fakeroot
		exclude="${exclude#/}"
		sedopts="$sedopts -e '\\|^${exclude%/}/|d'"
		sedopts="$sedopts -e '\\|^${exclude%/}\$|d'"
	done
	if [ -z "$lock" ]; then lock="$src"; fi
	log_exec_locked "$lock" "(cd $src; find . $find_opts | sed $sedopts | cpio $cpio_opts '$dst' 2>&1 | sed -e '/Invalid cross-device link/d')" \
	|	sed -n -e '\|/$|d' -e "s|^$dst/||p"
}

#
# Get the relative path from one absolute directory path to another 
#	Destination path is mandatory
#	Source path may be omitted and defaults to current directory
#	BOTH parameters must be ABSOLUTE DIRECTORY path names
#
get_relative_path() {
	local dst="${1%/}/"
	local src="${2:-$(pwd)}"
	src="${src%/}/"
	while [ "${src%%/*}" = "${dst%%/*}" ]; do
		src=${src#*/}
		dst=${dst#*/}
		if [ -z "$src" ]; then
			break
		fi
		if [ -z "$dst" ]; then
			# /A/B/C/D/E -> /A/B/C ==> ../..
			break
		fi
	done
	# /A/B/C -> /A/B/C/D/E ==> D/E
	# /A/B/C/D/E -> /A/B/C ==> ../..
	# /A/B/C/D/E -> /A/B/C/F/G ==> ../../F/G
	# /A/B/C/$src -> /A/B/C/$dst ==> ${upwards $src}/$dst
	while [ -n "$src" ]; do
		src=${src#*/}
		dst="../$dst"
	done
	dst=${dst%/}
	if [ -z "$dst" ]; then
		echo "."
	else
		echo "$dst"
	fi
}

#
# Get a relative path to go to the base of a given path
#
get_up_path() {
	local down_path=$1
	local up_path=
	for i in ${down_path//\// }; do
		up_path=$up_path/..
	done
	echo $up_path
}

#
# Simplify a path, suppressing .. and . references, as long as the referenced
# directory, is still the same
# 
simplify_path() {
	local path="$1/"
	path="${path//\/.\///}"
	path="${path//\/.\///}"
	path="${path//\/.\///}"
	path="${path//\/.\///}"
	path="${path//\/.\///}"
	path="${path//\/\///}"
	path="${path//\/\///}"
	path="${path//\/\///}"
	path="${path//\/\///}"
	path="${path//\/\///}"
	local start end base

	while [[ " $path" =~ " ../" ]]; do
		base="../$base"
		path=${path#../}
	done
	while true; do
		start=${path%%/../*}
		if [ "$start" = "$path/" ]; then break; fi
		if [ -z "$base$start" ]; then
			path=${path#/..} # /.. is /
			continue
		fi
		if [ ! -d "$base$start" ]; then break; fi
		end=${path#*/../}
		if [ "$base$start/.." -ef "${base%/}/${start%/*}" ]; then
			path="${start%/*}"/$end
		else
			break
		fi
	done
	echo $base$path | sed -e 's://*$::'
}

#
# Create a relative symbolic link
#
lnr() {
	local target=$1
	local symlink=$2
	local base=$(dirname $symlink)
	mkdir -p $base
	local rel=$(get_relative_path $target $base)
	ln -s $rel $symlink
}

######################################################################################################################
#                                                                                                                    #
#                                               Configuration options                                                #
#                                                                                                                    #
######################################################################################################################
#
# Platform options may be tested in various environments to select additional actions to perform; for example
# these may be used in patch options: patchname.patch@when=gcc>=8 is the name of a patch applied only if gcc version
# is greater or equal to 8
#
# Specific platform options will be derived from platform definition below
#
# Note that, for native and cross recipes, gcc will refer to the build host
# while all others will describe the target host; if you need to always
# refer the build GCC version, use native-gcc and target-gcc for the target
# GCC version.
#
declare -g -A platform_options # This script may be sourced in a function; without -g the variable would be local

######################################################################################################################
#                                                                                                                    #
#                                           PLATFORM and RECIPE management                                           #
#                                                                                                                    #
######################################################################################################################

#
# Package IDs are formed as follows:
#
#	name_vers-rev_arch
#
# where:
#	name    name of the package; it could contain '-' but no '.' or '_'
#	vers    version of the package; it could contain '.' but no '-' or '_'
#	rev     revision o fthe package build rules; it could contain '-' or '.' but no '_'
#
# Built binary package names follow the pattern
#
#	name-sub_vers-rev_arch
#
# where:
#	sub     sub-package; it could contain '.' but no '-' or '_'
#	arch    architecture for which the package was built; it could contain '-' but no '.' or '_'
# (others are as above)
#

######################################################################################################################
#                                                                                                                    #
#                                                 PLATFORM configuration                                             #
#                                                                                                                    #
######################################################################################################################

sw4l_platform_plugin=$(dir -d $sw4l_install_base/plugins/fr.ac6.platform_* | sort -V | tail -1)
if [ -z "$sw4l_platform_plugin" -a -d "$sw4l_install_base/plugins/fr.ac6.platform" ]; then
	# We should be in DEBUG mode, working from the plugins source directories, that have no version attached
	DEBUG_SW4L=true
	sw4l_platform_plugin=$sw4l_install_base/plugins/fr.ac6.platform
fi

# Platform-specific variables
platform_loc=$PLATFORM_LOC
platformDir=$platform_loc
platform_id=$(basename $platform_loc)
platform_name=${platform_id%%_*}
platform_version=${platform_id#*_}

# These variables should be provided by the platform...
: ${arch_compiler_prefix:=${TARGET_SYSROOT_LOC##*/}}
: ${generic_subpackages:=""}
: ${global_additional_subpackages:=""}

ac6_platform_manifest=$platform_loc/ac6_platform_manifest.xml
ac6_platform_env=$platform_loc/sysroots/$arch_compiler_prefix/etc/ac6_platform.env

get_glibc_version() {
	local gcc="$1"
	local glibc=$($gcc -print-file-name=libc.so.6 2> /dev/null)
	if [ -z "$glibc" -o ! -r "$glibc" ]; then return 0; fi
	glibc="$(readlink $glibc)"
	glibc="${glibc##*libc-}"
	glibc=${glibc%.so}
	echo "$glibc"
}

if [ "$ac6_platform_manifest" -nt "$ac6_platform_env"  \
-o   "$(current_script_file)" -nt "$ac6_platform_env" \
-o   ! -r "$ac6_platform_env" ]; then
	rm -rf "$ac6_platform_env" # Just to be sure...
	mkdir -p $(dirname "$ac6_platform_env")
	cat << EOF > $tmpdir/ac6_platform_env.sed
/<config langId="fr.ac6.platform.arch.target.settings">/,/<\/config>/{
s/^.*<property name="\([a-zA-Z0-9_]\+\)"[^>]*>\(..*\)<\/property>.*$/: \${\1:="\2"}; export \1/p
}
s/^.*<property name="keep_going"[^>]*>\(..*\)<\/property>.*$/: \${KEEP_GOING:="\1"}; export KEEP_GOING/p
EOF
	sed -n -f $tmpdir/ac6_platform_env.sed $ac6_platform_manifest > $ac6_platform_env
fi
if [ ! -r $ac6_platform_env ]; then
	echo "No platform found... Bail out!" >&2
	exit 1
fi
. $ac6_platform_env

package_compiler_version=${package_compiler_package#*_}; export package_compiler_version
kernel_compiler_version=${kernel_compiler_package#*_}; export kernel_compiler_version
if [ -z "$package_compiler_libc_version" ]; then
	package_compiler_libc_version=$(get_glibc_version $arch_compiler_prefix-gcc)
	if [ -n "$package_compiler_libc_version" ]; then
		echo "package_compiler_libc_version='$package_compiler_libc_version'; export package_compiler_libc_version" >> $ac6_platform_env
	fi
fi

# Now we must search for the machine and cpu definition files; it would be better if the IDE provide it...
ac6_machine_env=$platform_loc/sysroots/$arch_compiler_prefix/etc/ac6_machine.env
if [ -n "$machine" ]; then
	if [ -r $platform_loc/targets/machines/$machine ]; then
		sed -n -e 's/^\([^=][^=]*\)=\(.*$\)/machine_config_\1="\2"; export machine_config_\1/p' $platform_loc/targets/machines/$machine > $ac6_machine_env
	elif [ -r $sw4l_platform_plugin/resources/targets_def/machines/$machine ]; then
		sed -n -e 's/^\([^=][^=]*\)=\(.*$\)/machine_config_\1="\2"; export machine_config_\1/p' $sw4l_platform_plugin/resources/targets_def/machines/$machine > $ac6_machine_env
	else
		echo "Machine definition file for machine '$machine' not found"
		# Bypass error in debug mode (machine definitions are only in production plugins)
		if [ "$DEBUG_SW4L" != true ]; then
			exit 1
		else
			# Should never occur
			echo "Bypass error in debug mode"
		fi
	fi
	source $ac6_machine_env
fi
if [ -n "$machine_config_soc" ]; then
	if [ -r $platform_loc/targets/socs/$machine_config_soc ]; then
		sed -n -e 's/^\([^=][^=]*\)=\(.*$\)/soc_config_\1="\2"; export soc_config_\1/p' $platform_loc/targets/socs/$machine_config_soc >> $ac6_machine_env
	elif [ -r $sw4l_platform_plugin/resources/targets_def/socs/$machine_config_soc ]; then
		sed -n -e 's/^\([^=][^=]*\)=\(.*$\)/soc_config_\1="\2"; export soc_config_\1/p' $sw4l_platform_plugin/resources/targets_def/socs/$machine_config_soc >> $ac6_machine_env
	else
		echo "SOC definition file for SoC '$machine_config_soc' of machine '$machine' not found"
		# Bypass error in debug mode (machine definitions are only in production plugins)
		if [ "$DEBUG_SW4L" != true ]; then
			exit 1
		else
			# Should never occur
			echo "Bypass error in debug mode"
		fi
	fi
	source $ac6_machine_env
fi
if [ -n "$soc_config_cpu" -a -z "$machine_config_cpu" ]; then 
	export machine_config_cpu="$soc_config_cpu"
fi
if [ -n "$machine_config_cpu" ]; then
	if [ -r $platform_loc/targets/cpus/$machine_config_cpu ]; then
		sed -n -e 's/^\([^=][^=]*\)=\(.*$\)/cpu_config_\1="\2"; export cpu_config_\1/p' $platform_loc/targets/cpus/$machine_config_cpu >> $ac6_machine_env
	elif [ -r $sw4l_platform_plugin/resources/targets_def/cpus/$machine_config_cpu ]; then
		sed -n -e 's/^\([^=][^=]*\)=\(.*$\)/cpu_config_\1="\2"; export cpu_config_\1/p' $sw4l_platform_plugin/resources/targets_def/cpus/$machine_config_cpu >> $ac6_machine_env
	else
		echo "CPU definition file for cpu '$machine_config_cpu' of machine '$machine' not found"
		# Bypass error in debug mode (machine definitions are only in production plugins)
		if [ "$DEBUG_SW4L" != true ]; then
			exit 1
		else
			# Should never occur
			echo "Bypass error in debug mode"
		fi
	fi
	source $ac6_machine_env
fi
if [ -n "$cpu_config_generic" -a -z "$machine_config_arch" ]; then
	export machine_config_arch="$cpu_config_generic"
fi
if [ -n "$machine_config_arch" ]; then
	if [ -r $platform_loc/targets/arch/$machine_config_arch ]; then
		sed -n -e 's/^\([^=][^=]*\)=\(.*$\)/cpu_config_\1="\2"; export cpu_config_\1/p' $platform_loc/targets/arch/$machine_config_arch >> $ac6_machine_env
	elif [ -r $sw4l_platform_plugin/resources/targets_def/arch/$machine_config_arch ]; then
		sed -n -e 's/^\([^=][^=]*\)=\(.*$\)/cpu_config_\1="\2"; export cpu_config_\1/p' $sw4l_platform_plugin/resources/targets_def/arch/$machine_config_arch >> $ac6_machine_env
	else
		echo "Genric ARCH definition file for cpu '$machine_config_cpu' of machine '$machine' not found"
		# Bypass error in debug mode (machine definitions are only in production plugins)
		if [ "$DEBUG_SW4L" != true ]; then
			exit 1
		else
			# Should never occur
			echo "Bypass error in debug mode"
		fi
	fi
	source $ac6_machine_env
fi
machine_vars="machine_config_cpu machine_config_arch $(sed -n -e 's/^\([a-zA-Z0-9_]\+\)=.*$/\1/p' < $ac6_machine_env)"

# Define default target configuration options
platform_options[arch]="$arch_name"
if [[ "$arch_compiler_prefix" =~ "neon" ]]; then platform_options[neon]=true; fi
platform_options[target-gcc]=${package_compiler_version%%-*}
platform_options[target-glibc]=${package_compiler_libc_version}
platform_options[kgcc]=${kernel_compiler_version%%-*}
platform_options[machine]="$machine"
platform_options[vendor]="$machine_config_vendor"
platform_options[cpu]="$machine_config_cpu"
platform_options[soc]="$machine_config_soc"
platform_options[endian]="$cpu_config_endian"
platform_options[bits]="$cpu_config_bits"

ac6_build_env=$platform_loc/sysroots/native/etc/ac6_build.env
if [ "$ac6_platform_manifest" -nt "$ac6_build_env"  \
-o   "$(current_script_file)" -nt "$ac6_build_env" ]; then
	mkdir -p $(dirname $ac6_build_env)
	rm -f $ac6_build_env
	tmpbase=$tmpdir/ac6_build_env
	cat << EOF > $tmpbase.sed
/<config langId="fr.ac6.platform.arch.native.settings">/,/<\/config>/{
s/^.*<property name="\([a-zA-Z0-9_]\+\)"[^>]*>\(..*\)<\/property>.*$/: \${native_\1:="\2"}; export native_\1/p
}
EOF
	sed -n -f $tmpbase.sed $ac6_platform_manifest > $ac6_build_env
	native_lib_dlsearch_path=
	native_library_search_path=$(gcc -print-search-dirs | sed -n -e 's/^libraries:[[:space:]]*=*//p')
	for dir in ${native_library_search_path/:/ }; do
		dir=$(readlink -m $dir)
		if ! [[ " $native_lib_dlsearch_path " =~ " $dir " ]]; then
			native_lib_dlsearch_path="$native_lib_dlsearch_path $dir"
		fi
	done
fi
native_build_alias=
source $ac6_build_env
# Here we just need the native GCC version; can't cache this, as this may change dynamically
# We must use gcc as long as native-gcc is not yet installed...
if native-gcc --version > /dev/null 2>&1; then
	platform_options[native-gcc]=$(native-gcc --version | sed -n -e '/^/s/^gcc[[:space:]].*[[:space:]]//p' 2> /dev/null)
	platform_options[native-glibc]=$(get_glibc_version native-gcc)
else
	platform_options[native-gcc]=$(gcc --version | sed -n -e '/^/s/^gcc[[:space:]].*[[:space:]]//p' 2> /dev/null)
	platform_options[native-glibc]=$(get_glibc_version gcc)
fi

if [ -z "$native_build_alias" ]; then
	# Only trust the config.guess script installed by gnu-config-native, that should be
	# executed really soon when building a new platform
	if [ -r "$NATIVE_SYSROOT_LOC/usr/share/gnu-config/config.guess" ]; then
		native_build_alias=$(bash $NATIVE_SYSROOT_LOC/usr/share/gnu-config/config.guess)
		echo "export native_build_alias=\"$native_build_alias\"" >> $ac6_build_env
	elif [ -r "$WEAK_SYSROOT_LOC/usr/share/gnu-config/config.guess" ]; then
		native_build_alias=$(bash $WEAK_SYSROOT_LOC/usr/share/gnu-config/config.guess)
		echo "export native_build_alias=\"$native_build_alias\"" >> $ac6_build_env
	fi
	# Values provided by other scripts may be less safe, don't save them
	if [ -z "$native_build_alias" ]; then
		if [ -r config.guess ]; then
			native_build_alias=$(bash config.guess)
		fi
	fi
	if [ -z "$native_build_alias" ]; then
		if [ -r build-aux/config.guess ]; then
			native_build_alias=$(bash build-aux/config.guess)
		fi
	fi
	if [ -z "$native_build_alias" ]; then
		native_build_alias="x86_64-pc-linux-gnu"
	fi
fi

######################################################################################################################
#                                                                                                                    #
#                                                Dynamic CLASS management                                            #
#                                                                                                                    #
######################################################################################################################

# Here we can declare additional class inheritance
# Ideally this should be in the CLASS_utils subfile, but we cannot yet search these as we need this information
# to search them. In the future we should have a function add_interitance to dynamically add inheritance
# relationships that could be handled dynamically by the various search functions (as loading a class script may
# add new relationships... Probably this would need to have class definition files loaded transitively early on
# that could create this relationship here, implementing a kind of tsort (or using tsort itself).
declare -g -A inherited_classes # This script may be sourced in a function; without -g the variable would be local
inherited_classes["autotools"]="makefile"
inherited_classes["module"]="kernel"
inherited_classes["kernel"]="makefile"
inherited_classes["cpan"]="makefile"
inherited_classes["makefile"]="make"  # make is not really a class, just a synonym for makefile

#
# Some class items are in a directory different from platform.CLASS; provide these directories in
# the script_class_directory array
#
declare -g -A script_class_directory # This script may be sourced in a function; without -g the variable would be local

# Rootfs and utils are in directories linux.CLASS instead of platform.CLASS
script_class_directory["rootfs"]="linux.rootfs"
script_class_directory["utils"]="linux.utils"

# The base class is in linux.utils
script_class_directory["base"]="linux.utils"

# The default class items are in platform
script_class_directory["default"]="platform"

# Distutils3 items are in the distutils class directory
script_class_directory["distutils3"]="platform.distutils"

# The module class is in plugin platform.modules...
script_class_directory["module"]="platform.modules"

# These are the predefined classes; update this list when predefined classes are transformed in dynamic classes
# Note this variable must have leading and trailing spaces...
all_defined_classes=" autotools base cmake cpan distutils distutils3 kernel makefile meson module prebuilt rootfs scons secure "
all_dynamic_classes=
dbg dynclass "all_defined_classes='$all_defined_classes'"

# Dynamic classes are classes that are only defined by a template and scripts provided in a generic 
# dynamic classes plugin that, at least initially, is the classes subdirectory in linuc.utils
dynamic_classes_plugin="linux.utils/classes"
dbg dynclass "dynamic_classes_plugin='$dynamic_classes_plugin'"

# Search for the most recent dynamic classes plugin directory
dynamic_classes_plugin_dir=
for dir in "$sw4l_install_base/plugins/fr.ac6.${dynamic_classes_plugin%%/*}_"*/resources/scripts; do
	if [ -d "$dir" ]; then dynamic_classes_plugin_dir=$dir; fi
done
dbg dynclass "dynamic_classes_plugin_dir='$dynamic_classes_plugin_dir'"
dynamic_classes_plugin_subdir=${dynamic_classes_plugin#*/}
if [ "$dynamic_classes_plugin_subdir" = "dynamic_classes_plugin" ]; then
	dynamic_classes_plugin_subdir=.
fi
dbg dynclass "dynamic_classes_plugin_subdir='$dynamic_classes_plugin_subdir'"

script_class_dir() {
	local class="${1:-default}"
	local classdir=${script_class_directory[$class]:-platform.$class}
	echo $classdir
}

define_new_class() { # Inheritance and class directory should be derived from a class definition file
	local class="$1"
	local inherit="$2"
	local class_dir="$3"
	dbg dynclass "define_new_class '$class' '$inherit' '$class_dir'"
	
	if [[ " $all_defined_classes " =~ " $class " ]]; then return; fi
	if [ -n "$inherit" ]; then
		inherited_classes["$class"]="$inherit"
	fi
	if [ -n "$class_dir" ]; then
		script_class_directory["$class"]="$class_dir"
	else
		script_class_directory["$class"]="$class"
	fi
	all_defined_classes="$all_defined_classes$class "
	all_dynamic_classes="$all_dynamic_classes$class "
}

define_new_dynamic_classes() {
	local dir class class_name class_dir class_file class_inherit class_ext="class.tpl"
	local base_dir base_dirs="$platform_loc/classes \
	                          $platform_loc/scripts/plugins/$dynamic_classes_plugin \
	                          $dynamic_classes_plugin_dir/$dynamic_classes_plugin_subdir"
	(
		for base_dir in $base_dirs; do
			for class_file in $base_dir/*.$class_ext; do
				if [ -r $class_file ]; then
					basename $class_file .$class_ext
				fi
			done
			for class_dir in $base_dir/*; do
				if [ -d $class_dir ]; then
					basename $class_dir
				fi
			done
		done
	) | sort | uniq > $tmpdir/dyn_classes
	dbg_cat dynclass "$tmpdir/dyn_classes"

	cat <<- EOF > $tmpdir/class_reader.sed
		s/^[[:space:]*]*@class_dir:[^=]*=[[:space:]]*"\([^"][^"]*\)".*$/class_dir="\1"/p
		s/^[[:space:]*]*@inherit:[^=]*=[[:space:]]*"\([^"][^"]*\)".*$/class_inherit="\1"/p
	EOF
	dbg_cat dynclass $tmpdir/class_reader.sed
	for class in $(cat $tmpdir/dyn_classes); do
		class_file=
		class_name=$class
		class_inherit=
		for base_dir in $base_dirs; do
			if [ -r $base_dir/$class/$class.$class_ext ]; then
				class_file=$base_dir/$class/$class.$class_ext
				dbg_cat dynclass $class_file
				if [ -r $base_dir/$class.$class_ext ]; then
					warning "Class definition file found in both the class directory and the classes main directory"
					warning "Only the file in the class subdirectory ($class/$class.$class_ext) will be used"
					dbg_cat dynclass $base_dir/$class.$class_ext
				fi
				break
			fi
			if [ -r $base_dir/$class.$class_ext ]; then
				class_file=$base_dir/$class.$class_ext
				dbg_cat dynclass $class_file
				break
			fi
		done
		if [ -n "$class_file" -a -r "$class_file" ]; then 
			# Extract info from a class definition file
			sed -n -f $tmpdir/class_reader.sed < $class_file > $tmpdir/class_$class_name.sh
			dbg_cat dynclass $tmpdir/class_$class_name.sh
			source $tmpdir/class_$class_name.sh
		fi
		: ${class_dir:=$class}
		define_new_class "$class_name" "$class_inherit" "$class_dir"
	done
}
define_new_dynamic_classes
dbg dynclass "all_defined_classes='$all_defined_classes'"
dbg dynclass "all_dynamic_classes='$all_dynamic_classes'"

######################################################################################################################
#                                                                                                                    #
#                                                  RECIPE configuration                                              #
#                                                                                                                    #
######################################################################################################################

# Backward compatibility settings
: ${RECIPE_ID:=$PACKAGE_ID} 
: ${RECIPE_NAME:=$PACKAGE_NAME}
: ${RECIPE_VERSION:=$PACKAGE_VERSION}
: ${RECIPE_REV:=$PACKAGE_REV}

# Recipe specific variables
export recipe_id=$RECIPE_ID # Can't use directly as version/revision not at the proper place for sub-packages
export recipe_name=$RECIPE_NAME
export recipe_base_name=$recipe_name
export recipe_version=$RECIPE_VERSION
export recipe_revision=$RECIPE_REV

recipe_spec=$recipe_name
if [ -n "$recipe_version" ]; then
	recipe_spec="${recipe_spec}_${recipe_version}"
fi

if [ -z "$recipe_name" ]; then 
	# We execute a platform-specific task, just note it and provide defaults so that recipe-specific code works
	recipe_id=$platform_id
	recipe_name=$platform_name
	recipe_base_name=$recipe_name
	recipe_version=$platform_version
	recipe_is_platform=true
else
	recipe_is_platform=false
fi

# Default values
: ${recipe_revision:=r0}
: ${recipe_statedir:=.ac6_build}

# Recipe-specific derived variables
# First remove the type (native or cross) then the flavor (intermediate, initial,
# minimal or dummy); only one of each group can be present but who cares...
for extension in native cross intermediate initial minimal dummy; do
	recipe_base_name=${recipe_base_name%-$extension}
done
recipe_base_spec=$recipe_base_name
if [ -n "$recipe_version" ]; then
	recipe_base_spec="${recipe_base_spec}_${recipe_version}"
fi

export recipe_version_revision=${recipe_version}-${recipe_revision}
if $recipe_is_platform; then
	recipe_workdir="$tmpdir/platform_workdir"
else
	recipe_workdir="$taskBuildStatusDir"
fi
export recipe_workdir
#recipe_workdir="$WORKSPACE_LOC/$recipe_id" # Only valid after the Import buildstep // TODO: export from the IDE

# Create and source the bash-compatible recipe manifest
#
# This defines various variables that should be exported by the IDE:
#	recipe_manifest                 recipe manifest file
#	recipe_metadatadir              recipe metadata directory
#	recipe_kind                     package, kernel or rootfs 
#	recipe_is_native                true or false
#	recipe_is_cross	                true or false (a native package with target-specific data)
#	recipe_is_dummy                 true or false
#	recipe_is_autotooled            true or false
#	recipe_classes                  list of all recipe classes (c, autotools, compiler...)
#	recipe_buildtime_dependencies   list of all build-time dependencies
#	recipe_runtime_dependencies	    list of all run-time dependencies
# All these variables are exported to the environment
#	
# FIXME: export these from the IDE
#
if $recipe_is_platform; then
	recipe_manifest=
else
	recipe_manifest=$RECIPE_MANIFEST
	if [ -z "$recipe_manifest" ]; then
		for Xrecipe_kind in packages kernels rootfs; do
			Xrecipe_metadatadir="$platform_loc/metadata/$Xrecipe_kind/$recipe_id"
			Xrecipe_manifest="$Xrecipe_metadatadir/ac6_recipe_manifest.xml"
			if [ -r "$Xrecipe_manifest" ]; then
				recipe_manifest="$Xrecipe_manifest"
				break;
			fi
			Xrecipe_manifest="$Xrecipe_metadatadir/ac6_package_manifest.xml"
			if [ -r "$Xrecipe_manifest" ]; then
				recipe_manifest="$Xrecipe_manifest"
				break;
			fi
		done
		if [ "$Xrecipe_kind" != rootfs ]; then
			Xrecipe_kind=${Xrecipe_kind%s}
		fi
	fi
	if [ -n "$recipe_manifest" ]; then
		recipe_metadatadir=$(dirname "$recipe_manifest")
	fi
	if [ -d "$PROJECT_LOC" ]; then
		mkdir -p "$PROJECT_LOC/_build/.ac6_build"
		recipe_env="$PROJECT_LOC/_build/.ac6_build/$recipe_id.env"
		recipe_vars="$PROJECT_LOC/_build/.ac6_build/$recipe_id.vars"
		recipe_arch="$PROJECT_LOC/_build/.ac6_build/$recipe_id.arch-level"
	else
		# We are in Fetch, before PROJECT_LOC creation, and we should not create it, as Import may choke at it...
		recipe_env="$tmpdir/recipe_manifest.env"
		recipe_vars="$tmpdir/recipe_manifest.vars"
		recipe_arch="$tmpdir/recipe_manifest.arch-level"
	fi
fi

is_in_class() {
	local class="$1"
	if [[ " $recipe_classes " =~ " $class " ]]; then
		return 0
	else
		return 1
	fi
}

do_add_recipe_class()  {
	local recipe_classes="$1"
	shift
	local class inherit base_dir
	local global_base_dirs="$dynamic_classes_plugin_dir/$dynamic_classes_plugin_subdir \
	                        $platform_loc/scripts/plugins/$dynamic_classes_plugin \
	                        $platform_loc/classes"
	local platform_base_dirs="$platform_loc/classes"
	local recipe_base_dirs="$recipe_metadatadir/classes"
	local buildstep_base_dir="$recipe_metadatadir/buildsteps"
	
	for class; do
		if [[ " $recipe_classes " =~ " $class " ]]; then continue; fi
		# All class main template files has already been analyzed and provided global info like inheritance
		for inherit in ${inherited_classes["$class"]}; do
			recipe_classes=$(do_add_recipe_class "$recipe_classes" $inherit)
		done
		recipe_classes="$class $recipe_classes"
		
		#
		# Read and export the class values generated by the class templates or provided statically by the class
		# itself. More variables can be set early by the class in a $class.class.sh file, before recipe_utils
		# creates most of the recipe- and buildstep-specific variables; note that these variables may be overriden
		# or (more rarely) modified by recipe_utils.sh, so if you want to modify generated variables before the
		# buildsteps are executed, define them in the ${class}_utils.sh file that is sourced at the end of
		# execution of recipe_utils.sh, often using prepend_to_shell_variable or append_to_shell_variable.
		#
		# Only those in the recipe itself or in the platform classes directory can be generated by a template
		# (most of those by the main template, $class-$class.tpl, and edited in recipe properties are placed in
		# the recipe metadata, a few may be defined by a template named $class.platform.tpl and provided in
		# the platform/class directory).
		#
		# Buildstep-specific variables may be provided by a $class.$buildstep.tpl template, providing an
		# additionaltab in the properties (probably rarely used) or by marking the variables as buildstep-specicic
		# in the main template (more common).
		#
		# Global static values may be provided by a script placed in the plugin directories.
		#
		# These value defining files are in fact shell script that may define or compute variable values; they are
		# responsible of exporting them as needed. These files should be part of the buildstep hashes; that's why there
		# are buildstep-specific files.
		#
		for base_dir in $global_base_dirs; do
			if [ -r $base_dir/$class/$class.class.sh ]; then
				source $base_dir/$class/$class.class.sh
			fi
			if [ -r $base_dir/$class.class.sh ]; then
				source $base_dir/$class.class.sh
			fi
		done
		for base_dir in $platform_base_dirs; do
			if [ -r $base_dir/$class/$class.platform.vals ]; then
				source $base_dir/$class/$class.platform.vals
			fi
			if [ -r $base_dir/$class.platform.vals ]; then
				source $base_dir/$class.platform.vals
			fi
		done
		for base_dir in $recipe_base_dirs; do
			if [ -r $base_dir/$class/$class.recipe.vals ]; then
				source $base_dir/$class/$class.recipe.vals
			fi
			if [ -r $base_dir/$class.recipe.vals ]; then
				source $base_dir/$class.recipe.vals
			fi
		done
		for base_dir in $buildstep_base_dirs; do
			if [ -r $base_dir/$class/$class.$current_buildstep.vals ]; then
				source $base_dir/$class/$class.$current_buildstep.vals
			fi
			if [ -r $base_dir/$class.$current_buildstep.vals ]; then
				source $base_dir/$class.$current_buildstep.vals
			fi
		done
	done
	echo $recipe_classes
}

if [ -n "$recipe_manifest" ]; then
	export recipe_manifest
	export recipe_metadatadir
	if [ ! -r "$recipe_env" ] || \
	   [ "$recipe_manifest" -nt "$recipe_env" ] || \
	   [ "$current_buildstep_script" -nt "$recipe_env" ] || \
	   [ "$(current_script_file)" -nt "$recipe_env" ]; then
		rm -f $recipe_env
		tmpbase=$tmpdir/recipe_manifest-utils

		cat << EOF > $tmpbase.sed
/^.*<typeId>fr\.ac6\.platform\..*<\/typeId>.*$/{
s/^.*<typeId>fr\.ac6\.platform\.\(.*\)<\/typeId>.*$/\1/
s/.*\.//g
p
}
s#.*<property.*name="build_classes".*>\(.*\)</property>#\1#p
EOF
		# All classes are in the base class, which must be given first, as we
		# build the class list from the most generic to the most specific, so
		# that class values from more specific classes can override/augment
		# values from less specific classes)...
		# recipe_classes are sorted from the most specific to the most generic
		# as do_add_recipe_class adds classes in front of the passed list
		Xrecipe_classes= 
		Xrecipe_classes_reversed=
		for i in "base" $(sed -n -f $tmpbase.sed $recipe_manifest); do
			Xrecipe_classes=$(do_add_recipe_class "$Xrecipe_classes" $i)
		done
		for i in $Xrecipe_classes; do
			Xrecipe_classes_reversed="$i $Xrecipe_classes_reversed"
		done
		Xrecipe_classes=" ${Xrecipe_classes} " # spaces are needed for matching class names
		if [[ " $Xrecipe_classes " =~ " autotools " ]]; then
			Xrecipe_is_autotooled=true
		else
			Xrecipe_is_autotooled=false
		fi

		cat << EOF > $tmpbase.sed
s/^.*<dependency.*type="runtime".*>\(.*\)<\/dependency>.*$/\1/p
EOF
		Xrecipe_runtime_dependencies=
		for i in $(sed -n -f $tmpbase.sed $recipe_manifest); do
			Xrecipe_runtime_dependencies="$Xrecipe_runtime_dependencies $i"
		done
		
		cat << EOF > $tmpbase.sed
s/^.*<dependency.*equals="\([^"]*\)".*type="buildtime".*>\(.*\)<\/dependency>.*$/\2_\1/p
s/^.*<dependency.*type="buildtime".*equals="\([^"]*\)".*>\(.*\)<\/dependency>.*$/\1_\2/p
s/^.*<dependency.*type="buildtime".*>\(.*\)<\/dependency>.*$/\1/p
EOF
		Xrecipe_buildtime_dependencies=
		for i in $(sed -n -f $tmpbase.sed $recipe_manifest); do
			Xrecipe_buildtime_dependencies="$Xrecipe_buildtime_dependencies $i"
		done
		
		Xrecipe_is_native=false
		if grep -q '<properties.*isNative="true".*>' $recipe_manifest; then
			Xrecipe_is_native=true
		fi
		
		Xrecipe_is_cross=false
		if $Xrecipe_is_native; then
			if [ "${recipe_name##*-}" = "cross" ]; then
				Xrecipe_is_cross=true
			fi
		fi
		
		Xrecipe_is_dummy=false
		if [ "recipe_is_dummy" = "true" ] \
		|| [[ "${recipe_name} " =~ "-dummy-" ]] \
		|| [[ "${recipe_name} " =~ "-dummy " ]] \
		|| [[ "${recipe_name} " =~ "-minimal-" ]] \
		|| [[ "${recipe_name} " =~ "-minimal " ]] \
		|| [[ "${recipe_name} " =~ "-initial-" ]] \
		|| [[ "${recipe_name} " =~ "-initial " ]] \
		|| [[ "${recipe_name} " =~ "-intermediate-" ]] \
		|| [[ "${recipe_name} " =~ "-intermediate " ]]; then
			if $Xrecipe_is_native; then
				# Dummy recipes work only with the weak sysroot
				Xrecipe_is_dummy=true
			else
				fatal 98 "Only native recipes can generate dummy or intermediate output"
			fi
		fi
		
		for i in kind is_native is_cross is_dummy is_autotooled \
		         classes classes_reversed \
		         runtime_dependencies buildtime_dependencies; do
			j=recipe_$i
			k=X$j
			echo "$j=\"${!k}\"; export $j"
			unset $k
		done                                                                       >   $recipe_env
		source $recipe_env

		# Base directory where all work is done
		workDir=${PROJECT_LOC%/}
		echo "workDir=\"\${PROJECT_LOC%/}\"; export workDir"                       >>  $recipe_env

		# Directory where source files are located
		: ${src_loc:=$workDir} 
		: ${srcDir:=$src_loc}
		# Always get the srcDir from the recipe manifest (and ignore it when getting all other variables)
		if grep -s -q '<property name="srcDir"' $recipe_manifest; then
			srcDir=$(sed -n -e 's!^.*<property name="srcDir".*>\(.*\)</property>.*$!\1!p' $recipe_manifest)
		fi
		# srcDir may reference workDir in deferred expansion mode, so expand now
		eval srcDir=$srcDir
		
		if [ -z "$srcDir" ]; then
			subDir=
		elif ! [[ " $srcDir" =~ " /" ]]; then 
			subDir=/$srcDir
		else
			subDir=${srcDir#$workDir}
			if [ "$subDir" = "$srcDir" ]; then
				subDir=$(basename $workDir)
				subDir=${srcDir##/*/$subDir}
				if [ "$subDir" = "$srcDir" ]; then
					# If workspace has changed, we keep the provided subDir but
					# here srcDir is out of the project... Come back at project root
					# This should never happens but try to be safe anyway
					subDir=
				fi
			fi
		fi
		srcDir=${workDir}$subDir
		echo "srcDir=\"\${workDir}$subDir\"; export srcDir"                         >>  $recipe_env

		# Directory where out-of-source package compile step place files
		# The IDE also uses this directory for stamp files. 		
		provided_build_loc="$build_loc"
		if [ "$BUILD_IN_SOURCE_TREE" = "true" ]; then
			build_loc="$srcDir"
		elif [ "$BUILD_IN_SOURCE_TREE" = "false" ]; then
			build_loc="$workDir/_build"
		elif [ -n "$USER_BUILD_DIR" ]; then
			case "$USER_BUILD_DIR" in
				\$*)# Use paths prefixed by some SW4L directory variable as-is
					build_loc="$USER_BUILD_DIR"
					;;
				srcDir/* | workDir/*)
					# Use working directory names as variable names
					var=${USER_BUILD_DIR%%/*}
					build_loc=${!var}/${USER_BUILD_DIR#*/}
					;;
				/*) # Use absolute path names as-is too
					build_loc="$USER_BUILD_DIR"
					;;
				*)	# Place relative path names in the work directory
					build_loc="$workDir/$USER_BUILD_DIR"
					;;
			esac
		elif is_in_class autotools \
		||   is_in_class cmake \
		||   is_in_class meson; then
			build_loc="$workDir/_build" # Place to use for build files
		elif is_in_class module \
		||   is_in_class kernel; then
			build_loc="$srcDir" # Place to use for build files
		else
			build_loc="$srcDir" # Place to use for build files
		fi
		: ${buildDir:=$build_loc}
#		if [ "$provided_build_loc" != "$build_loc" ]; then
#			echo "WARNING: IDE set build location at '$provided_build_loc' while build system set it at '$buildDir'" >&2
#		fi
		subDir=${buildDir#$workDir}
		echo "buildDir=\"\${workDir}$subDir\"; export buildDir"                     >>  $recipe_env
		
		# Directories where package install step place files
		: ${install_loc:=$workDir/_install} # Place to use for target files
		: ${installDir:=$install_loc}
		subDir=${installDir#$workDir}
		echo "installDir=\"\${workDir}$subDir\"; export installDir"                 >>  $recipe_env
		
		: ${native_install_loc:=$workDir/_sysroot} # Place to use for native files
		: ${nativeInstallDir:=$native_install_loc}
		native_install_loc="$nativeInstallDir"
		subDir=${nativeInstallDir#$workDir}
		echo "nativeInstallDir=\"\${workDir}$subDir\"; export nativeInstallDir"     >>  $recipe_env
		
		# The only variable we really need here is configure_other_flags; it should be passed by the IDE
		# but it seems it's incorrectly passed so get it from the manifest and ignore it later...
		# Furthermore ensure that variables are NOT expanded here, they are not yet defined...
		# We do that by including the value in single quotes AND escaping dollars twice, while
		# protecting embedded single quotes
		sed -n -e 's!^.*<property name="configure_other_flags".*>\(.*\)</property>.*$!\1!p' $recipe_manifest | \
		if true; then # Single $ escape is enough to be equivalent to yocto
			sed -e "s/'/'\"'\"'/g" \
			    -e 's/\([^$]\)\$/\1\\$/g' \
			    -e 's/^\$/\\$/' \
			    -e 's/^.*$/'"configure_other_flags='&'/" \
			    >> $recipe_env
		else # But double escape defers variable expansion after configure has run
			sed -e "s/'/'\"'\"'/g" \
			    -e 's/\\\$/\\\\&/g' \
			    -e 's/\([^$]\)\$/\1\\\\\\$/g' \
			    -e 's/^\$/\\\\\\$/' \
			    -e 's/^.*$/'"configure_other_flags='&'/" \
			    >> $recipe_env
		fi
		
		# Here we should check if the recipe use the platform settings (for compiler, machine, etc.)
		# or has its own definition, so that we may not redefine compiler_prefix and compiler_path
		# (and maybe other platform definition variables) when not appropriate.
	fi
elif ! $recipe_is_platform; then
	fatal 1 "Can't find recipe manifest for '$recipe_id' in platform '$(basename $platform_loc)'"
fi
#echo "-------------------8<------------------- $recipe_env" >&2
#cat $recipe_env >&2
#echo "-------------------8<------------------- $recipe_env" >&2

if ! $recipe_is_platform; then
	if [ -r $recipe_env ]; then
		source $recipe_env
		src_loc="$srcDir"
		build_loc="$buildDir"
		install_loc="$installDir"
		native_install_loc="$nativeInstallDir"
		unset Xrecipe_kind
	else
		fatal 2 "No recipe environment definition file"
	fi
	
	export REAL_NATIVE_SYSROOT_LOC="$NATIVE_SYSROOT_LOC"
	if $recipe_is_dummy; then
		# Dummy recipes work first with the weak sysroot, so reverse native and weak in the PATH
		PATH="$(prefix_path $ORIGINAL_PATH $NATIVE_SYSROOT_LOC $WEAK_SYSROOT_LOC)"
		NATIVE_SYSROOT_LOC="$WEAK_SYSROOT_LOC"
		SYSROOT_LOC="$WEAK_SYSROOT_LOC"
	fi
	
	if $recipe_is_native; then
		platform_options[gcc]=${platform_options[native-gcc]}
		platform_options[glibc]=${platform_options[native-glibc]}
	else
		platform_options[gcc]=${platform_options[target-gcc]}
		platform_options[glibc]=${platform_options[target-glibc]}
	fi
fi

add_recipe_class()  {
	recipe_classes=$(do_add_recipe_class "$recipe_classes" "$@")
}

#
# Load profile fragments, now that we know from which native sysroot load them
#
profileDir=${NATIVE_SYSROOT_LOC}/etc/profile.d
if [ -d ${profileDir} ]; then
	for frag in ${profileDir}/*; do
		source $frag
	done
	unset frag
fi

#
# Yocto-compatible names, for cases where we may import yocto scripts
# 	Warning: you should avoid using these as much as possible, these
#	definitions are just there for the occasional case where you 
#	overlooked one of their uses...
#
S=$srcDir
B=$buildDir
D=$installDir
WORKDIR=$workDir
PN="$recipe_name"
BPN="$recipe_base_name"
PV="$recipe_version"
PR="$recipe_revision"

# Also reset IDE-provided variables
src_loc=$srcDir
build_loc=$buildDir
install_loc=$installDir
native_install_loc=$nativeInstallDir

#
# Define target characteristics (for searching target-specific files)
# 
# Note we can't export all BUILD_XXX, HOST_XXX and TARGET_XXX values
# right from here as some of those, like TARGET_ARCH, are used in make
# default rules and would then crash makefile execution...
#
compute_machine_characteristics() {
	local alias_parser=$1
	local typ=$2
	local TYP=$3
	local cpu=$4
	
	if [ -z "$alias_parser" ]; then return; fi
	
	local recipe_arch=${alias_parser%%-*}  		        # x86_64, arm, aarch64
	alias_parser=${alias_parser#*-}
	local recipe_machine=${alias_parser%%-*}	        # pc, cortexa7hf_vfpv4_neon
	local recipe_os=${alias_parser#*-}      	        # linux, linux-gnueabi, ...

	export recipe_${typ}_arch=$recipe_arch      		# x86_64, arm, aarch64
	export recipe_${typ}_machine=$recipe_machine		# pc, cortexa7hf_vfpv4_neon
	export recipe_${typ}_os=$recipe_os          		# linux, linux-gnueabi, ...
	
	local _config_arch=${cpu}_config_arch
	local config_arch=${!_config_arch}
	if [ -z "$config_arch" ]; then
		config_arch=${recipe_arch}
		case $recipe_arch in
			arm)                                    ;;
			armeb)          config_arch=arm         ;;
			aarch64)        config_arch=arm         ;;
			aarch64_be)     config_arch=arm         ;;
			i?86)           config_arch=ix86        ;;
	        x86_64)         config_arch=ix86        ;;
			mips)                                   ;;
			mips64)         config_arch=mips        ;;
			mips64el)       config_arch=mips        ;;
			mipsisa64r6)    config_arch=mips        ;;
			mipsisa64r6el)  config_arch=mips        ;;
			mipsel)         config_arch=mips        ;;
			mipsisa32r6)    config_arch=mips        ;;
			mipsisa32r6el)  config_arch=mips        ;;
	        powerpc)                                ;;
	        powerpc64)      config_arch=powerpc     ;;
	        ppc)            config_arch=powerpc     ;;
	        ppc64)          config_arch=powerpc     ;;
	        ppc64le)        config_arch=powerpc     ;;
	        sparc)                                  ;;
	        riscv32)        config_arch=riscv       ;;
	        riscv64)        config_arch=riscv       ;;
	        nios2)          config_arch=nios2       ;;
	        microblaze)                             ;;
	        microblazeeb)   config_arch=microblaze  ;;
	        microblazeel)   config_arch=microblaze  ;;
	
			*)
				echo "architecture $recipe_arch unsupported" >&2
				exit 1
				;;
		esac
		export $_config_arch=${config_arch}
		echo "$_config_arch=$config_arch; export $_config_arch" >> $recipe_env
	fi
	eval ${TYP}_ARCH="$config_arch"
	
	local _config_bits=${cpu}_config_bits
	local config_bits=${!_config_bits}
	if [ -z "$config_bits" ]; then
		# this should be provided by the cpu file
		config_bits=32
		case $recipe_arch in
			arm)                                    ;;
			armeb)                                  ;;
			aarch64)        config_bits=64          ;;
			aarch64_be)     config_bits=64          ;;
			i?86)                                   ;;
	        x86_64)         config_bits=64          ;;
			mips)                                   ;;
			mips64)         config_bits=64          ;;
			mips64el)       config_bits=64          ;;
			mipsisa64r6)    config_bits=64          ;;
			mipsisa64r6el)  config_bits=64          ;;
			mipsel)                                 ;;
			mipsisa32r6)                            ;;
			mipsisa32r6el)                          ;;
	        powerpc)                                ;;
	        powerpc64)      config_bits=64          ;;
	        ppc)                                    ;;
	        ppc64)          config_bits=64          ;;
	        ppc64le)        config_bits=64          ;;
	        sparc)                                  ;;
	        riscv32)                                ;;
	        riscv64)        config_bits=64          ;;
	        nios2)                                  ;;
	        microblaze)                             ;;
	        microblazeeb)                           ;;
	        microblazeel)                           ;;
			
			*)
				echo "word size unknown for architecture $recipe_arch" >&2
				exit 2
				;;
		esac
		export $_config_bits=$config_bits
		echo "$_config_bits=$config_bits; export $_config_bits" >> $recipe_env
	fi
	eval ${TYP}_BITS="$config_bits"
	
	local _config_endian=${cpu}_config_endian
	local config_endian=${!_config_endian}
	if [ -z "$config_endian" ]; then
		config_endian="little"
		case $recipe_arch in
			arm)                                    ;;
			armeb)          config_endian=big       ;;
			aarch64)                                ;;
			aarch64_be)     config_endian=big       ;;
			i?86)                                   ;;
	        x86_64)                                 ;;
			mips)           config_endian=big       ;;
			mips64)  		config_endian=big       ;;
			mips64el)                               ;;
			mipsel)                                 ;;
			mipsisa64r6)    config_endian=big       ;;
			mipsisa32r6)    config_endian=big       ;;
	        powerpc)        config_endian=big       ;;
	        powerpc64)      config_endian=big       ;;
	        ppc)            config_endian=big       ;;
	        ppc64)          config_endian=big       ;;
	        ppc64le)                                ;;
	        sparc)          config_endian=big       ;;
	        riscv32)                                ;;
	        riscv64)                                ;;
	        microblaze)     config_endian=big       ;;
	        microblazeeb)   config_endian=big       ;;
	        microblazeel)                           ;;
	        nios2)                                  ;;
			
			*)
				echo "endianness unknown for architecture $recipe_arch" >&2
				exit 3
				;;
		esac
		export $_config_endian=$config_endian
		echo "$_config_endian=$config_endian; export $_config_endian" >> $recipe_env
	fi
	case $config_endian in
		little) 
			eval ${TYP}_ENDIANNESS="le"
			;;
		big)
			eval ${TYP}_ENDIANNESS="be"
			;;
		*)
			echo "Unknown CPU endianness '$config_endian'; should be 'little' or 'big'" >&2
			exit 4
			;;
	esac
	
	local _config_os=${cpu}_config_os
	local config_os=${!_config_os}
	if [ -z "$config_os" ]; then
		config_os=$recipe_os
		case $recipe_os in
			darwin9)	config_os=darwin            ;;
			linux)                                  ;;
			linux-*)	config_os=linux             ;;
			uclinux-*)  config_os=uclinux	        ;;
			cygwin)                                 ;;
			mingw32)    config_os=mingw             ;;
			
			*)
				echo "Unknown OS '$recipe_os'" >&2
				exit 5
				;;
		esac
		export $_config_os=$config_os
		echo "$_config_os=$config_os; export $_config_os" >> $recipe_env
	fi
	eval ${TYP}_OS=$config_os
	
	local _config_libc=${cpu}_config_libc
	local config_libc=${!_config_libc}
	if [ -z "$config_libc" ]; then
		config_libc=glibc
		case $recipe_os in
			darwin9)        config_libc=            ;;
			linux)                                  ;;
			linux-gnu)                              ;;
	        linux-musl)     config_libc=musl        ;;
	        linux-muslx32)  config_libc=musl        ;;
	        linux-musleabi) config_libc=musl        ;;
	        linux-muslspe)  config_libc=musl        ;;
			linux-*)                                ;;
			uclinux-uclibc) config_libc=uclibc	    ;;
			cygwin)                                 ;;
			mingw32)                                ;;
			
			*)
				echo "Library unknown for OS '$target_os'" >&2
				exit 6
				;;
		esac
		export $_config_libc=$config_libc
		echo "$_config_libc=$config_libc; export $_config_libc" >> $recipe_env
	fi
	eval ${TYP}_LIBC=$cpu_config_libc
	
	export ${typ}_config_machine=$recipe_machine
	eval ${TYP}_MACHINE=$recipe_machine
	eval ${TYP}_OS_VARIANT=$recipe_os
}

if ! $recipe_is_platform; then
	compute_machine_characteristics "${native_build_alias}" build BUILD bld
	if $recipe_is_cross; then
		# Target defaults to host
		compute_machine_characteristics "${target_compiler_prefix}" target TARGET tgt
		compute_machine_characteristics "${arch_compiler_prefix}" host HOST cpu
	elif $recipe_is_native; then
		# Host and target are build
		compute_machine_characteristics "${native_build_alias}" host HOST cpu
		compute_machine_characteristics "${native_build_alias}" target TARGET cpu
	else
		# Target is host
		target_compiler_prefix=$arch_compiler_prefix
		compute_machine_characteristics "${target_compiler_prefix}" target TARGET tgt
		compute_machine_characteristics "${arch_compiler_prefix}" host HOST cpu
	fi
fi

#
# Package architecture definition.
#	Architectures are placed in an array, indexed by the precision of the architecture
#		0 = portable		(arch = noarch)
#		1 = generic 		(arch = CPU architecture: arm arm64, ...)
#		2 = CPU type  		(arch = CPU type: cortexa7, ...)
#		3 = CPU variant  	(arch = CPU: cortexa7hf-vfpv4-neon, ...)
#		4 = vendor  		(arch = Vendor name: st nxp, ...)
#		5 = SoC-family		(arch = Soc family: i.MX, STM32, ...)
#		6 = SoC subfamily	(arch = SoC kind: i.MX8, MPSoC, ...)
#		7 = SoC         	(arch = SoC name: i.MX8mq, STM32MP15, ...)
#		8 = Machine			(arch = machine: sabrelite, avenger96, ...)
#		9 = Platform		(arch = platform name)
#
# This script may be sourced in a function; without -g the variable would be local
declare -g -a arch_level_name=(
	[0]="noarch"
	[1]="$arch_name"
	[2]="${machine_config_cpu%%-*}"
	[3]="$machine_config_cpu"
	[4]="$machine_config_vendor"
	[5]="$machine_config_family"	# To be defined
	[6]="$machine_config_variant"	# To be defined
	[7]="$machine_config_soc"
	[8]="$machine"
	[9]="$platform_name"
)

#
# Get the package architecture name
#
package_arch_level=1
package_arch_name() {
	local level=$package_arch_level
	local name="${arch_level_name[level]}"
	while [ -z "$name" -a $level -lt 9 ]; do
		name="${arch_level_name[level++]}"
	done
	echo "$name"
}

#
# Adds an architecture variant
#
declare -g -a arch_variants # This script may be sourced in a function; without -g the variable would be local
add_arch_variant() {
	local level="$1"
	shift
	local variants="${arch_variants[level]}"
	arch_variants[level]="$variants $*"
}

record_recipe_kind() {
	add_arch_variant 10 "$*"
}

#
# Search a file in a target-dependant subdirectory.
#	Returns the full file path
#	sets the package_architecture variable to the most precise value
#	sets also the package_arch_prio to the precision of the architecture:
#
machine_specific_file() {
	local baseDir="$1"
	local fileName="$2"
	local file level name prefix=.
	
	# Seach the most specific, that is the last found...
	for level in "${!arch_level_name[@]}"; do
		for name in $prefix ${arch_level_name[level]} ${arch_variants[level]}; do
			if [ -r "$baseDir/$name/$fileName" ]; then
				if [ "$name" = "." ]; then
					file="$baseDir/$fileName"
				else
					file="$baseDir/$name/$fileName"
				fi
				file_level=$level
			fi
		done
		prefix=
	done
	
	# raise the arch level of the recipe if needed (but only if less than 9)
	if [ $level -le 9 -a $level -gt $package_arch_level ]; then
		package_arch_level=$level
	fi
	
	# and return the found file name
	echo $file
}

#
# Provide standard defaults for installation directories
#
# Note these are the same as $autotools_dirs, but in an order
# better suited for displaying them to the user
#
# This script may be sourced in a function; without -g the variable would be local
declare -g -a install_directories=(
	prefix
	exec_prefix
	bindir
	base_bindir
	sbindir
	base_sbindir
	libexecdir
	libdir
	base_libdir
	pkgconfigdir
	includedir
	datarootdir
	datadir
	sysconfdir
	sharedstatedir
	localstatedir
	runstatedir
	localedir
	mandir
	infodir
	docdir
	htmldir
	dvidir
	pdfdir
	psdir
	systemd_system_unitdir
	systemd_unitdir
	systemd_user_unitdir
	nonarch_base_libdir
	nonarch_libdir
	servicedir
)

default_installation_directories() {
	local hbase="$1"
	local tbase="$2"
	# Provide sensible defaults for installation directories
	# If prefix is not provided, use /usr
	# If prefix is provided as / use /usr as prefix for includedir and datarootdir
	# If prefix is provided as /usr, use /etc as sysconfdir 
	# These will be passed to the configure script for autotooled packages,
	# others will find these arguments in additional_configure_opts
	: ${prefix:=$tbase/usr}
	prefix="${prefix%/}"
	base_prefix="${prefix%/usr}"
	# beware that exec_prefix may use unexpanded references to prefix
	eval exec_prefix="$prefix"
	: ${exec_prefix:=$prefix}
	exec_prefix="${exec_prefix%/}"
	base_exec_prefix="${exec_prefix%/usr}"
	if $recipe_is_cross; then
		# Place executables in native sysroot 
		exec_prefix="${hbase%/}${exec_prefix#${tbase%/}}"
		base_exec_prefix="${hbase%/}${base_exec_prefix#${tbase%/}}"
		if [ "$PLACE_LIBRARIES_ON_TARGET" = "true" ]; then
			# In rare cases libraries must go in target sysroot, as if cross-compiled,
			# and include files should be generated in/taken from the target sysroot
			: ${libdir:='$prefix/lib'}
			: ${base_libdir:='$base_prefix/lib'}
		else # This is the default
			# As libraries are native libraries, includes must be native too
			: ${includedir:='$base_exec_prefix/usr/include'}
			: ${oldincludedir:='$base_exec_prefix/usr/include'}
		fi
	fi

	# Set default package installation directories in installDir based on prefix
	# These are mostly the official default values but tailored to fit the
	# yocto setup; when we need to use another value, either it was specified
	# in configure_other_flags or it was provided by setup_configure_options
	# and passed to configure
	: ${bindir:='$exec_prefix/bin'}
	: ${base_bindir:='$base_exec_prefix/bin'}
	: ${sbindir:='$exec_prefix/sbin'}
	: ${base_sbindir:='$base_exec_prefix/sbin'}
	: ${libexecdir:='$exec_prefix/libexec'}
	: ${libdir:='$exec_prefix/lib'}
	: ${base_libdir:='$base_exec_prefix/lib'}
	: ${datarootdir:='$base_prefix/usr/share'}
	: ${datadir:='$datarootdir'}
	: ${sysconfdir:='$base_prefix/etc'}
	: ${pkgconfigdir:='$base_prefix/usr/lib/pkgconfig'}
	: ${sharedstatedir:='$base_prefix/com'}
	: ${localstatedir:='$base_prefix/var'}
	: ${runstatedir:='$localstatedir/run'}
	: ${includedir:='$base_prefix/usr/include'}
	: ${oldincludedir:='$base_prefix/usr/include'}
	: ${localedir:='$datarootdir/locale'}
	: ${infodir:='$datarootdir/info'}
	: ${docdir:='$datarootdir/doc'}
	: ${htmldir:='$docdir/html'}
	: ${dvidir:='$docdir/dvi'}
	: ${pdfdir:='$docdir/pdf'}
	: ${psdir:='$docdir/ps'}
	: ${mandir:='$datarootdir/man'}
	: ${systemd_system_unitdir:='$base_libdir/systemd/system'}
	: ${systemd_unitdir:='$base_libdir/systemd'}
	: ${systemd_user_unitdir:='$libdir/systemd/user'}
	: ${nonarch_base_libdir:='$base_libdir'}
	: ${nonarch_libdir:='$libdir'}
	: ${servicedir:='$base_prefix/srv'}
	
	# Now expand and export all installation directories
	local dir
	for dir in "${install_directories[@]}"; do
		eval "$dir=\"${!dir%/}\""
		export $dir
	done

	# Ensure prefix and exec_prefix are not empty
	: ${prefix:=/}
	: ${exec_prefix:=/}
}

if ! $recipe_is_platform; then
	if $recipe_is_native; then record_recipe_kind native; fi
	if $recipe_is_cross; then record_recipe_kind cross; fi
fi

#
# Analyze autotools prefix and install directory specifications
# so that we can use them in our scripts
# Here we also ensure we follow the proper populate sysroot conventions:
# 	"native" 	packages are configured and installed in the native sysroot
#	"cross" 	packages are configured and installed with binaries in native 
#				sysroot and data in target sysroot
#	"target" 	packages are configured to execute on the target and 
#				populated in the target sysroot
# Native and cross packages execute from the native sysroot, using files
# coming natively from either the native or target sysroots
#
# This routine will normalize all configuration paths and update 
# configure_other_flags , except if $DO_NOT_UPDATE_INSTALL_DIRECTORIES is
# set to true
#

configure_auto_dirs="prefix exec_prefix \
					 bindir sbindir libexecdir sysconfdir \
					 sharedstatedir localstatedir libdir \
					 includedir oldincludedir \
					 datarootdir datadir infodir  mandir"
other_auto_dirs="localedir docdir htmldir dvidir pdfdir psdir"
if [ "$DO_NOT_AUTORECONF" != true ]; then
	configure_auto_dirs="$configure_auto_dirs $other_auto_dirs"
fi
autotools_dirs="$configure_auto_dirs  $other_auto_dirs \
                base_bindir base_sbindir base_libdir \
                runstatedir pkgconfigdir systemd_system_unitdir \
                systemd_unitdir systemd_user_unitdir \
                nonarch_base_libdir nonarch_libdir \
                servicedir
                "
default_configure_options="--disable-silent-rules"

# This script may be sourced in a function; without -g these variables would be local
declare -g -a configure_flags=() original_configure_flags=() updated_configure_flags=()
setup_configure_options() {
	# Native and cross packages must be configured to execute in the
	# platform-provided native sysroot
	local cbase # common base directory, in which both tbase and hbase are located
	local hbase # main base directory, files that must be on the host
	local tbase # target base, for files that should be on the target
	# --prefix should be on the target (tbase)
	#	except that --bindir, --sbindir and --libexec MUST be on host (hbase)
	# --exec_prefix (so rootdatadir and sysconfdir) should be on target (tbase)
	local opt dir var base 
	local tgt_flags bdir_flags dir_flags dflt_flags cfg_flags
	local need_libtool_sysroot=false
	
	if [ "$DO_NOT_UPDATE_INSTALL_DIRECTORIES" != "true" ]; then
		if $recipe_is_native; then
			hbase="${NATIVE_SYSROOT_LOC%/}"
			if $recipe_is_cross; then
				tbase="${TARGET_SYSROOT_LOC%/}"
			else
				tbase="${hbase}"
			fi
			cbase=$(dirname $hbase)
		else
			cbase=""
			hbase=""
			tbase=""
		fi
	else
		cbase=""
		hbase=""
		tbase=""
	fi
	
	# Define default maintainer mode (maybe overriden by user options)
	local enable_maintainer_mode
	if $recipe_is_autotooled; then
		local enable_maintainer_mode=no
	fi
	
	#
	# If we do not regenerate configure script it's better not to enable
	# maintainer mode as it will probably fail, if regenerating does
	# (which is probably why we do not regenerate them)
	#
	# If we do regenerate configure scripts, maintainer mode is usually
	# useless as, by definition, the configure script are up-to-date;
	# furthermore some files are manually copied because autopoint may
	# overwrite them with older versions, so enabling maintainer mode
	# may again break them...
	#
	# However, of course, if the user really need maintainer mode, he
	# can enable it explicitly by adding --enable-maintainer-mode to
	# configure_other_flags
	#
	#if [ "$DO_NOT_AUTORECONF" = "true" ]; then
	#	enable_maintainer_mode=no
	#else
	#	enable_maintainer_mode=yes
	#fi
	specify_maintainer_mode=false
	local passthru=false
	declare -a cfg_flags dir_flags tgt_flags bdir_flags dflt_flags
	# Here we must be careful NOT to expand variables, as they may not yet be defined
	mapfile -t original_configure_flags < <(split_in_words "$configure_other_flags")

	for opt in "${original_configure_flags[@]}"; do
		if $passthru; then
			cfg_flags+=("$opt")
			continue
		fi
		case $opt in
			--)	# Remaining flags must be passed as-is
				passthru=true
				cfg_flags+=("$opt")
				continue
				;;
			--enable-maintainer-mode=*)
				enable_maintainer_mode=${opt#--*=}
				specify_maintainer_mode=true
				;;
			--enable-maintainer-mode)
				enable_maintainer_mode=yes
				specify_maintainer_mode=true
				;;
			--disable-maintainer-mode)
				enable_maintainer_mode=no
				specify_maintainer_mode=true
				;;
			--*=*)
				dir=${opt#--*=}
				opt=${opt#--}
				opt=${opt%%=*}
				base=""
				var="${opt//-/_}"
				case $opt in
					with-libtool-sysroot)
						need_libtool_sysroot=true;
						;;
					with-* | without-* | enable-* | disable-*)
						;;
					exec_prefix | bindir | sbindir | libexecdir)
						base="$hbase"
						;;
					prefix)
						base="$tbase"
						;;
					*dir)
						base="$tbase"
						;;
					build | host | target)
						# Empty values or references to ${opt}_alias defaults
						# to references to configure_${opt}_alias
						if [ -z "$dir" -o "$dir" = "${opt}"'_alias}' ]; then 
							dir='${configure_'"${opt}"'_alias}'
						fi
						var="${opt}_alias"
						;;
				esac
				if [ -n "$base" ]; then
					if [[ ! " $dir" =~ " $cbase/" ]] && [[ ! "$dir" =~ "\$" ]]; then
						# Do not relocate paths that are in any sysroot, nor those that
						# are not yet resolved (or are empty or relative paths)
						case "$dir" in
							TARGET/*)
								base="$tbase"
								dir="${base%/}/${dir#TARGET/}"
								;;
							NATIVE/*)
								base="$hbase"
								dir="${base%/}/${dir#NATIVE/}"
								;;
							/*)
								dir="${base%/}/${dir#/}"
								;;
							*)
								base=""
								;;
						esac
					fi
				fi
				case $opt in
					with-* | without-* | enable-* | disable-*)
						eval $var='${dir//$/\$}' # Keep references unexpanded
						cfg_flags+=("--$opt=${dir//$/\$}")
						;;
					build | host | target)
						eval $var='${dir//$/\$}' # Keep references unexpanded
						tgt_flags+=("--$opt=${dir//$/\$}")
						;;
					*)	# These are directory options; only read their values if they are
						# explicit, ignoring the value and replacing it by a variable
						# reference if they are empty (used to also skip references)
						# if [ -n "$dir" ] && [[ ! "$dir" =~ "\$" ]]; then
						if [ -n "$dir" ]; then
							eval $var="$dir"
						fi
						if ! $recipe_is_autotooled \
						|| ! [[ " $configure_auto_dirs " =~ " $var " ]]; then
							if [ -z "$configure_non_auto_dirs" ] \
							|| ! [[ " $configure_non_auto_dirs " =~ " $var " ]]; then
								# default directories will be added later
								if [[ "$var " =~ "dir " ]] || [[ "$var " =~ "_alias " ]]; then
									dir_flags+=("--$opt=${dir//$/\$}")
								else
									cfg_flags+=("--$opt=${dir//$/\$}")
								fi
							fi
						fi
						;;
				esac
				;;

			--*)
				# Note that only long options are forwarded as booleans; We automatically set
				# with/without and enable/disable as a single with or enable variable set to true/false
				# We also transform no-foobar in foobar=false
				#
				# Warning: only configure options matching usual GNU conventions will be handled correctly
				# options introduced by single characters will NOT be parsed and exported as variables
				dir=${opt#--}
				var=$dir
				val=true
				case $dir in
					without-*)	var=with-${var#without-}; val=false ;;
					disable-*)	var=enable-${var#disable-}; val=false ;;
					no-*)		var=${var#no-}; val=false ;;
				esac
				eval configure_${dir//-/_}=$val
				cfg_flags+=("$opt")
				;;

			-*)
				cfg_flags+=("$opt")
				;;

			*=*\ *)
				var=${opt%%=*}
				val=${opt#*=}
				eval configure_${var//-/_}="$val"
				cfg_flags+=("$var='$val'")
				;;

			*=*)
				var=${opt%%=*}
				val=${opt#*=}
				eval configure_${var//-/_}="$val"
				cfg_flags+=("$opt")
				;;

			*\ *)
				cfg_flags+=("'$opt'")
				;;

			*)
				cfg_flags+=("$opt")
				;;
		esac
	done

	# Provide sensible defaults for build, host and target systems
	configure_build_alias="$native_build_alias"
	if [ -z "$build_alias" ]; then
		build_alias="$configure_build_alias"
		if $recipe_is_autotooled; then
			tgt_flags+=("--build=$configure_build_alias")
		fi
	fi
	# Ensure variable references are expanded here
	eval configure_build_alias="$configure_build_alias"
	eval build_alias="$build_alias"
	export configure_build_alias
	export build_alias
	
	if $recipe_is_native; then
		configure_host_alias="$configure_build_alias"
	else
		configure_host_alias="$arch_compiler_prefix"
	fi
	if [ -z "$host_alias" ]; then
		host_alias=$configure_host_alias
		if $recipe_is_autotooled; then
			tgt_flags+=("--host=$configure_host_alias")
		fi
	fi
	# Ensure variable references are expanded here
	eval configure_host_alias="$configure_host_alias"
	eval host_alias="$host_alias"
	export configure_host_alias
	export host_alias
	
	if $recipe_is_cross; then
		configure_target_alias="$arch_compiler_prefix"
	else
		configure_target_alias="$configure_host_alias"
	fi
	if [ -z "$target_alias" ]; then
		target_alias=$configure_target_alias
		# Only add a --target option for cross tools
		if $recipe_is_cross && $recipe_is_autotooled; then
			tgt_flags+=("--target=$configure_target_alias")
		fi
	fi
	# Ensure variable references are expanded here
	eval configure_target_alias="$configure_target_alias"
	eval target_alias="$target_alias"
	export configure_target_alias
	export target_alias
	
	# Propagate maintainer mode
	if $specify_maintainer_mode; then
		if [ "$enable_maintainer_mode" = yes ]; then
			tgt_flags+=("--enable-maintainer-mode")
		else
			tgt_flags+=("--disable-maintainer-mode")
		fi
	else
		if [ "$enable_maintainer_mode" = yes ]; then
			dflt_flags+=("--enable-maintainer-mode")
		else
			true # dflt_flags+=("--disable-maintainer-mode")
		fi
	fi

	# Apply defaults to installation directories
	default_installation_directories "$hbase" "$tbase"
	
	for opt in $default_configure_options; do
		dflt_flags+=("$opt")
	done
	
	# Note that cfg_flags MUST remain last as they may include an end-of-options marker (--)
	if $recipe_is_autotooled; then
		if ! $need_libtool_sysroot; then
			# Here we should check if libtool is used by looking at an LT_INIT autoconf macro?
			# For now just consider it's always used and hop it does not crash at configure-time
			need_libtool_sysroot=true
		fi
		if $need_libtool_sysroot; then
			dir_flags+=( "--with-libtool-sysroot=\$SYSROOT_LOC" )
		fi
		for dir in $configure_auto_dirs; do
			if ! [[ " $configure_non_auto_dirs " =~ " $dir " ]]; then
				bdir_flags+=("--$dir=\"\${$dir}\"")
			fi
		done
		updated_configure_flags=("${tgt_flags[@]}" "${bdir_flags[@]}" "${dir_flags[@]}" "${dflt_flags[@]}" "${cfg_flags[@]}")
	else
		updated_configure_flags=("${tgt_flags[@]}" "${dir_flags[@]}" "${cfg_flags[@]}")
	fi
}

#expand variables in the build configuration flags
expand_sw4l_build_configuration() {
	local idx
	for idx in "${!updated_configure_flags[@]}"; do
		updated_configure_flags[idx]="$(eval requote_arg ${updated_configure_flags[idx]})"
	done
#set -vx
#echo "configure_host_alias='${configure_host_alias}'" >&2
#	local idx opt cfg_opts=$tmpdir/config.opts
#	for idx in "${!updated_configure_flags[@]}"; do
#		echo "updated_configure_flags[$idx]=\"${updated_configure_flags[$idx]}\""
#	done > $cfg_opts
#set +vx
#echo "-------------------8<------------------- $cfg_opts" >&2
#cat $cfg_opts >&2
#echo "-------------------8<------------------- $cfg_opts" >&2
#set -vx
#	source $cfg_opts
#set +vx
}

#echo native_build_alias="$native_build_alias" >&2
#echo arch_compiler_prefix="$arch_compiler_prefix" >&2
#echo "recipe_is_native='$recipe_is_native'" >&2
#echo "recipe_is_cross='$recipe_is_cross'" >&2

if ! $recipe_is_platform; then
	setup_configure_options
	
	if [ -n "$configure_build_alias" ]; then
		export configure_build_prefix=${configure_build_alias}-
	else
		export configure_build_prefix=
	fi
	
	if [ -n "$configure_host_alias" ]; then
		export configure_host_prefix=${configure_host_alias}-
	else
		export configure_host_prefix=
	fi
	
	if [ -n "$configure_target_alias" ]; then
		export configure_target_prefix=${configure_target_alias}-
	else
		export configure_target_prefix=
	fi
	
	export BUILD_CPP=${BUILD_CPP:=${configure_build_prefix}gcc -E}
	export BUILD_CC=${BUILD_CC:=${configure_build_prefix}gcc}
	export BUILD_CXX=${BUILD_CXX:=${configure_build_prefix}g++}
	export BUILD_LD=${BUILD_LD:=${configure_build_prefix}ld}
	export BUILD_CCLD=${BUILD_CCLD:=${configure_build_prefix}gcc}
	export BUILD_CXXLD=${BUILD_CXXLD:=${configure_build_prefix}g++}
	export BUILD_NM=${BUILD_NM:=${configure_build_prefix}nm}
	export BUILD_AR=${BUILD_AR:=${configure_build_prefix}ar}
	export BUILD_RANLIB=${BUILD_RANLIB:=${configure_build_prefix}ranlib}
	export BUILD_AS=${BUILD_AS:=${configure_build_prefix}as}
	export BUILD_READELF=${BUILD_READELF:=${configure_build_prefix}readelf}
	export BUILD_STRIP=${BUILD_STRIP:=${configure_build_prefix}strip}
	export BUILD_STRINGS=${BUILD_STRINGS:=${configure_build_prefix}strings}
	export BUILD_OBJCOPY=${BUILD_OBJCOPY:=${configure_build_prefix}objcopy}
	export BUILD_OBJDUMP=${BUILD_OBJDUMP:=${configure_build_prefix}objdump}
	
	if ! $recipe_is_autotooled; then
		export CPP=${CPP:=${configure_host_prefix}gcc -E}
		export CC=${CC:=${configure_host_prefix}gcc}
		export CXX=${CXX:=${configure_host_prefix}g++}
		export LD=${LD:=${configure_host_prefix}ld}
		export CCLD=${CCLD:=${configure_host_prefix}gcc}
		export CXXLD=${CXXLD:=${configure_host_prefix}g++}
		export NM=${NM:=${configure_host_prefix}nm}
		export AR=${AR:=${configure_host_prefix}ar}
		export RANLIB=${RANLIB:=${configure_host_prefix}ranlib}
		export AS=${AS:=${configure_host_prefix}as}
		export READELF=${READELF:=${configure_host_prefix}readelf}
		export STRIP=${STRIP:=${configure_host_prefix}strip}
		export STRINGS=${STRINGS:=${configure_host_prefix}strings}
		export OBJCOPY=${OBJCOPY:=${configure_host_prefix}objcopy}
		export OBJDUMP=${OBJDUMP:=${configure_host_prefix}objdump}
	fi
	
	#echo "configure_other_flags='$configure_other_flags'" >&2
	#echo native_build_alias="$native_build_alias"
	#echo arch_compiler_prefix="$arch_compiler_prefix"
	#echo configure_build_alias="$configure_build_alias"
	#echo configure_host_alias="$configure_host_alias"
	#echo configure_target_alias="$configure_target_alias"
	
	export TARGET_CFLAGS=${TARGET_CFLAGS:="-O2 -pipe"}
	export BUILD_CFLAGS=${BUILD_CFLAGS:="-O2 -pipe"}
	if ! $recipe_is_native; then
		export CFLAGS=${CFLAGS:=$TARGET_CFLAGS}
	else
		export CFLAGS=${CFLAGS:=$BUILD_CFLAGS}
	fi
	
	#echo "TARGET_CFLAGS='$TARGET_CFLAGS'"
	#echo "BUILD_CFLAGS='$BUILD_CFLAGS'"
	#echo "CFLAGS='$CFLAGS'"
	
	export TARGET_CXXFLAGS=${TARGET_CXXFLAGS:=$TARGET_CFLAGS}
	export BUILD_CXXFLAGS=${BUILD_CXXFLAGS:=$BUILD_CFLAGS}
	if ! $recipe_is_native; then
		export CXXFLAGS=${CXXFLAGS:=$TARGET_CXXFLAGS}
	else
		export CXXFLAGS=${CXXFLAGS:=$BUILD_CXXFLAGS}
	fi
	
	#echo "TARGET_CXXFLAGS='$TARGET_CXXFLAGS'"
	#echo "BUILD_CXXFLAGS='$BUILD_CXXFLAGS'"
	#echo "CXXFLAGS='$CXXFLAGS'"	
	
	export TARGET_CPPFLAGS=${TARGET_CPPFLAGS:=""}
	export BUILD_CPPFLAGS=${BUILD_CPPFLAGS:=""}
	if ! $recipe_is_native; then
		export CPPFLAGS=${CPPFLAGS:=$TARGET_CPPFLAGS}
	else
		export CPPFLAGS=${CPPFLAGS:=$BUILD_CPPFLAGS}
	fi
	
	#echo "TARGET_CPPFLAGS='$TARGET_CPPFLAGS'"
	#echo "BUILD_CPPFLAGS='$BUILD_CPPFLAGS'"
	#echo "CPPFLAGS='$CPPFLAGS'"	
	
	export TARGET_LDFLAGS=${TARGET_LDFLAGS:=""}
	export BUILD_LDFLAGS=${BUILD_LDFLAGS:=""}
	if ! $recipe_is_native; then
		export LDFLAGS=${LDFLAGS:=$TARGET_LDFLAGS}
	else
		export LDFLAGS=${LDFLAGS:=$BUILD_LDFLAGS}
	fi
	
	#echo "TARGET_LDFLAGS='$TARGET_LDFLAGS'"
	#echo "BUILD_LDFLAGS='$BUILD_LDFLAGS'"
	#echo "LDFLAGS='$LDFLAGS'"	
	
	# And ensure all configure options are fully expanded 
	expand_sw4l_build_configuration
	
	#
	# Now we must get all variables defined by the recipe (some may
	# not have been placed in the environment by the IDE). Furthermore
	# we must expand the configure variables that may be referenced in 
	# these so that they are properly defined; that's why this must have
	# been postponed after the configure options have been analyzed and
	# updated and the derived values defined.
	#
	if [ -n "$recipe_manifest" ]; then
		if [ ! -r "$recipe_vars" ] || \
		   [ "$recipe_manifest" -nt "$recipe_vars" ] || \
		   [ "$current_buildstep_script" -nt "$recipe_vars" ] || \
		   [ "$(current_script_file)" -nt "$recipe_vars" ]; then
			rm -f $recipe_vars
			tmpbase=$tmpdir/recipe_vars-utils
			# Force the definition of environment variables defined as properties
			# so that embedded variables are expanded
			cat > $tmpbase-1.sed << EOF
/<properties.*>/,/<\\/properties>/{
s/^.*<property name="\\([a-zA-Z0-9_]\\+\\)"[^>]*>\\(..*\\)<\\/property>.*$/\\1=\\2/p
}
EOF
			cat > $tmpbase-2.sed << EOF
/^srcDir=/d
/^configure_other_flags=/d
s/"/\\\"/g
s/\\\$/\\\$/g
s/^\\([^=][^=]*\\)=\\(.*\\)$/\\1=\"\\2\"; export \\1/
EOF
			sed -n -f $tmpbase-1.sed $recipe_manifest | sed -f $tmpbase-2.sed | tr '`' "'" > $recipe_vars
		fi
	fi
	
	# Finally apply the definitive configuration values (they may have been
	# reset by the values from the recipe manifest), variables expanded in them
	export configure_other_flags="${updated_configure_flags[*]}" # for backward compatibility
	export updated_configure_other_flags="$configure_other_flags" # for backward compatibility
	#echo "configure_other_flags='$configure_other_flags'" >&2
	export configure_flags=("${updated_configure_flags[@]}")
	
	if $recipe_is_autotooled; then
		#
		# Aclocal must search first in the current sysroot, when compiling for the target
		# so that it finds macros installed by target tools; we thus set ACLOCAL_PATH
		# here as it is useful for auto-reconfigure, configure and compile (that may
		# call autoconf/automake/aclocal).Regretfully there cannot be a cross-aclocal
		# so we have to set this in the environment...
		#
		if [ -r "$SYSROOT_LOC/usr/share/aclocal/libtool.m4" ]; then
			if [ -n "$ACLOCAL_PATH" ]; then
				export ACLOCAL_PATH="$ACLOCAL_PATH:$SYSROOT_LOC/usr/share/aclocal"
			else
				export ACLOCAL_PATH="$SYSROOT_LOC/usr/share/aclocal"
			fi
		fi
		if $recipe_is_native; then
			if [ -r "$WEAK_SYSROOT_LOC/usr/share/aclocal/libtool.m4" ]; then
				if [ -n "$ACLOCAL_PATH" ]; then
					export ACLOCAL_PATH="$ACLOCAL_PATH:$WEAK_SYSROOT_LOC/usr/share/aclocal"
				else
					export ACLOCAL_PATH="$WEAK_SYSROOT_LOC/usr/share/aclocal"
				fi
			fi
		else
			if [ -r "$NATIVE_SYSROOT_LOC/usr/share/aclocal/libtool.m4" ]; then
				if [ -n "$ACLOCAL_PATH" ]; then
					export ACLOCAL_PATH="$ACLOCAL_PATH:$NATIVE_SYSROOT_LOC/usr/share/aclocal"
				else
					export ACLOCAL_PATH="$NATIVE_SYSROOT_LOC/usr/share/aclocal"
				fi
			fi
		fi
		
		export lt_cv_sys_lib_dlsearch_path_spec="${libdir} ${base_libdir}"
		if $recipe_is_native; then
			lt_cv_sys_lib_dlsearch_path_spec="$lt_cv_sys_lib_dlsearch_path_spec $native_lib_dlsearch_path"
		fi
		
		# When building tools for use at build-time it's recommended for the build
		# system to use these variables when cross-compiling.
		# (http://sources.redhat.com/autobook/autobook/autobook_270.html)
		export CPP_FOR_BUILD="${BUILD_CPP}"
		export CPPFLAGS_FOR_BUILD="${BUILD_CPPFLAGS}"
		
		export CC_FOR_BUILD="${BUILD_CC}"
		export CFLAGS_FOR_BUILD="${BUILD_CFLAGS}"
		
		export CXX_FOR_BUILD="${BUILD_CXX}"
		export CXXFLAGS_FOR_BUILD="${BUILD_CXXFLAGS}"
		
		export LD_FOR_BUILD="${BUILD_LD}"
		export CCLD_FOR_BUILD="${BUILD_CCLD}"
		export CXXLD_FOR_BUILD="${BUILD_CXXLD}"
		export LDFLAGS_FOR_BUILD="${BUILD_LDFLAGS}"
		
	fi
	
	# Now that all configuration-dependent variables are set, read the other
	# recipe vars, that may reference them
	if [ -r $recipe_vars ]; then
		source $recipe_vars
	fi
	
	# And ensure all variables that must be exported are...
	for var in $additional_exported_vars; do
		export $var
	done
fi

# Do not forget to restore the package architecture level (and to save it if it has changed)
save_package_arch_level() {
	if [ -n "$recipe_arch" ]; then
		if [ $package_arch_level -gt $saved_package_arch_level ]; then
			echo "package_arch_level=$package_arch_level" > $recipe_arch
		fi
	fi
}

if [ -n "$recipe_arch" -a -r "$recipe_arch" ]; then
	source $recipe_arch
	saved_package_arch_level=$package_arch_level
	record_cleanup save_package_arch_level
fi

# Display the build configuration
display_sw4l_build_configuration() {
	echo ""
	echo "#####################################################################"
	echo "#                                                                   #"
	echo "#                     Recipe build configuration                    #"
	echo "#                                                                   #"
	echo "#####################################################################"
	echo ""
	echo "Recipe name:        $(trim $recipe_name)"
	echo "Recipe base name:   $(trim $recipe_base_name)"
	echo "Package version:    $(trim $recipe_version)"
	echo "Recipe revision:    $(trim $recipe_revision)"
	echo "Recipe ID:          $(trim $recipe_id)"
	echo "Recipe classes:     $(trim $recipe_classes)"
	echo ""
	if $recipe_is_native; then
		if $recipe_is_cross; then
			echo "Recipe type:        Cross"
		elif $recipe_is_dummy; then
			echo "Recipe type:        Intermediate"
		else
			echo "Recipe type:        Native"
		fi
	else
		echo "Recipe type:        Target"
	fi
	echo "Build:              $(trim $configure_build_alias)"
	echo "Host:               $(trim $configure_host_alias)"
	if $recipe_is_cross; then
		echo "Target:             $(trim $configure_target_alias)"
	fi
	echo ""
	echo "Source directory:   $(trim $srcDir)"
	echo "Build directory:    $(trim $buildDir)"
	echo "Install directory:  $(trim $installDir)"
	echo "Recipe sysroot:     $(trim $SYSROOT_LOC)"
	echo "Native sysroot:     $(trim $NATIVE_SYSROOT_LOC)"
	echo "Target sysroot:     $(trim $TARGET_SYSROOT_LOC)"
	echo ""
	echo "Installation directories (these are exported variables)"
	echo "======================================================="
	echo ""
	local dir
	local spaces="                 "
	for dir in "${install_directories[@]}"; do
		echo "	$dir=${spaces:${#dir}}${!dir}"
	done
	echo ""
	echo "Configuration options"
	echo "====================="
	echo ""
	local opt
	for opt in "${configure_flags[@]}"; do
		echo "	$opt"
	done
	echo ""
	echo "Environment variables"
	echo "====================="
	echo ""
	for var in configure_build_alias configure_host_alias configure_target_alias \
	           machine_config_vendor machine_config_cpu machine_config_kernel_defconfig \
	           machine_config_devicetree machine_config_soc machine_config_serial_consoles \
	           machine_config_serial_consoles $machine_vars \
	           cpu_config_cpu_generic cpu_config_cpu_gcc cpu_config_cpu cpu_config_arch \
	           cpu_config_bits cpu_config_os cpu_config_libc cpu_config_fpu cpu_config_float_abi \
	           cpu_config_cpu_architecture_variant cpu_config_cpu_architecture \
	           cpu_config_generic cpu_config_endian \
	           package_compiler_libc_version \
	           recipe_target_arch recipe_target_machine recipe_target_os \
	           BUILD_CPP BUILD_CPPFLAGS BUILD_CC BUILD_CFLAGS \
	           BUILD_CXX BUILD_CXXFLAGS BUILD_LD BUILD_LDFLAGS \
	           BUILD_NM BUILD_AR BUILD_READELF BUILD_STRIP \
	           TARGET_CPPFLAGS TARGET_CFLAGS TARGET_CXXFLAGS TARGET_LDFLAGS TARGET_OS \
	           HOST_ARCH HOST_BITS HOST_ENDIANNESS HOST_OS HOST_LIBC HOST_MACHINE \
	           CPP_FOR_BUILD CPPFLAGS_FOR_BUILD CC_FOR_BUILD CFLAGS_FOR_BUILD \
	           CXX_FOR_BUILD CXXFLAGS_FOR_BUILD LD_FOR_BUILD LDFLAGS_FOR_BUILD \
	           ACLOCAL_PATH lt_cv_sys_lib_dlsearch_path_spec \
	           CPP CPPFLAGS CC CFLAGS CXX CXXFLAGS LD LDFLAGS \
	           NM AR READELF STRIP $additional_exported_vars
	do
		echo $var
	done | sort | uniq | while read var; do
		if [ "${!var+set}" = "set" ]; then
			echo "	$var='${!var}'"
		fi
	done
	echo ""
	echo "#####################################################################"
	echo ""
}

if [ "$current_buildstep" = "configure" \
-o   "$current_buildstep" = "install" \
-o   "$current_buildstep" = "development-shell" ]; then
	display_sw4l_build_configuration
else
	# display_sw4l_build_configuration
	true
fi	
	
#
# The variables below are used for pkg-config PC file generation
# They also define which variables must be exported
#

# If a recipe adds new values there, it should NOT expand this list but add them
# to the non_standard_dir_vars; these must be listed in the order they are defined.
# Note that prefix is NOT in the list as it is usually the base for most others
install_dir_vars="bindir sbindir libdir base_bindir base_sbindir base_libdir \
                  libexecdir datarootdir datadir sysconfdir \
                  localstatedir sharedstatedir localedir \
                  includedir oldincludedir mandir infodir \
                  docdir htmldir dvidir pdfdir psdir \
                  systemd_system_unitdir systemd_unitdir \
                  systemd_user_unitdir nonarch_base_libdir \
                  nonarch_libdir servicedir"
non_standard_install_dir_vars=

# Only export install directories in native or cross recipes, not on target ones
# as they are not usable to access files...
if $recipe_is_native; then
	for dir_name in $install_dir_vars $non_standard_install_dir_vars; do
		eval export $dir_name="${!dir_name}"
	done
fi

# Finally we may need to access the readelf utility
call_readelf() {
	if [ -z "$executable_readelf" ]; then
		executable_readelf="${configure_host_prefix}readelf"
		if [ ! -x "$(which "$executable_readelf" 2> /dev/null)" ]; then
			if [ -n "${compiler_prefix}" ] && ! $recipe_is_native; then
		    	executable_readelf="${compiler_prefix}-readelf"
		    fi
			if [ ! -x "$(which "${executable_readelf}" 2> /dev/null)" ]; then
			    executable_readelf="readelf"
			fi
		fi
		if [ ! -x "$(which "$executable_readelf" 2> /dev/null )" ]; then
			executable_readelf=
		    errror "no suitable readelf program found"
		fi
	fi
	$executable_readelf "$@"
}

######################################################################################################################
#                                                                                                                    #
#                                                   Task configuration                                               #
#                                                                                                                    #
######################################################################################################################

: ${current_task_id=task.${current_task}.${recipe_id}}
case $current_task in
	populate-sysroot)
		current_task_statedir="${SYSROOT_LOC}/${recipe_statedir}"
		;;
	split | package)
		current_task_statedir="${PLATFORM_LOC}/feeds/${recipe_statedir}"
		;;
	fetch)
		current_task_statedir="${DOWNLOAD_LOC}/${recipe_statedir}"
		;;
	*)
		current_task_statedir="${recipe_workdir}/${recipe_statedir}"
		;;
esac
current_task_statefile="$current_task_statedir/${current_task_id}"

task_execute_date() {
	date -u +"%F %T %N"
}
task_start_date="$(task_execute_date)"

# Most tasks are executed from $buildDir
case $current_task in
	fetch | import)
		;;
	apply-patches | auto-reconfigure | autogen)
		cd $srcDir
		;;
	split | package)
		cd $installDir
		;;
	*)	
		if [ -d "$buildDir" ]; then
			cd $buildDir
		else
			cd $srcDir
		fi
		;;
esac

#
# Compatibility variable definitions
#
srcdir="$srcDir"

######################################################################################################################
#                                                                                                                    #
#                                                    Script utilities                                                #
#                                                                                                                    #
######################################################################################################################

#
# Search and source recipe-dependent scripts
#
# All these functions will search for a script in various directories:
#		The platform package (or kernel/platform/rootfs)-specific buildstep script directory
#		The platform generic script directory
#		The IDE default script library directory
# 	The script SCRIPT is searched (in order) as SCRIPT SCRIPT.sh do_SCRIPT do_SCRIPT.sh
#
# There are twelve functions in this group
#
#		search_script				search for the first script of a given name and outputs its filepath if found
#		search_all_scripts			search for all scripts of a given name and outputs their filepaths
#		load_script					source the provided script
#		load_script_log				source the provided script, logging that
#		load_script_once			source the provided script if not yet loaded 
#		source_script				search for the required script and sources it if found
#		source_script_log			search for the required script and sources it if found, logging that
#		source_script_once			search for the required script and sources it if found and not yet loaded
#		require_script				search for the required script and sources it but exit in error if not found
#		require_script_once			search for the required script and sources it if not yet loaded but exit in error if not found
#		source_all_scripts			source all scripts of a given name in search order
#		source_all_scripts_reverse	source all scripts of a given name in reversed search order
#		do_buildstep_prepend		search and execute a buildstep prepend scripts, if there is any, in search order
#		do_buildstep_append			search and execute a buildstep append scripts, if there is any, in reversed order
#
# In all buildstep scripts, the prepend scripts will be sourced after the environment was set up (variables, 
# working directories, etc.) but before any real work was done.
#
# The append scripts will be sourced at the real end of the buildstep script, just before its result is
# installed in the global directory if it's not a recipe-local buildstep
#
# All scripts sourced during a task are automatically hashed and the hash saved 
# (sorted) in an includes file so that changing a file may invalidate the task
#
if [ -d "${current_task_statedir}" ]; then
	loaded_scripts_state=${current_task_statedir}/${current_task_id}.includes
else # Project not yet created by Import, just place state file in some temporary directory
	loaded_scripts_state=$tmpdir/${current_task_id}.includes
fi
rm -f "$loaded_scripts_state"
touch "$loaded_scripts_state"

file_name_normalize=$tmpdir/${current_task_id}.file_name_normalize
record_cleanup rm -f "$file_name_normalize"
cat << EOF > $file_name_normalize
\|$sw4l_install_base/plugins/|{
s|$sw4l_install_base/plugins/|PLUGINS   |
n
}
\|$recipe_metadatadir|{
s|$recipe_metadatadir/|RECIPE    |
n
}
\|$PLATFORM_LOC|{
s|$PLATFORM_LOC/|PLATFORM  |
n
}
\|$buildDir|{
s|$buildDir/|BUILD     |
n
}
\|$PROJECT_LOC|{
s|$PROJECT_LOC/|PROJECT   |
n
}
\|$WORKSPACE_LOC|{
s|$WORKSPACE_LOC/|WORKSPACE |
n
}
s|/|ROOT      /|
EOF

already_loaded_scripts=
record_script_load() {
	local script_file="${1:-$(current_script_file 1)}"
# For now keep IDE-provided scripts in the checksums...
#	if [[ " $script_file " =~ " $sw4l_install_base/" ]]; then
#		return # Don't record scripts taken from the IDE itself, they don't change
#	fi

	# Only record a script once, even if loaded several times
	if [[ " $already_loaded_scripts " =~ " $script_file " ]]; then
		# Script was already loaded, so recorded, just return 1 so that caller knows that
		return 1
	fi
	already_loaded_scripts="$already_loaded_scripts $script_file"
	touch "$loaded_scripts_state"
	md5sum "$script_file" | sed -f $file_name_normalize >> "$loaded_scripts_state"
	return 0
}
record_script_load $(current_script_file 2)	# This is the main script (may be omitted as already tested)
record_script_load $(current_script_file 1) # This is the "utils.sh" script
record_script_load $log_utils_script_file	# This is the utils.sh-sourced "log_utils.sh" script
record_script_load $(current_script_file)	# This is the current "recipe_utils.sh" script

setup_scripts_dir() {
	local dir subdir classdir lastchance="$script_lib_dir" 
	local class classdir all_classes dyn_classes static_classes all_classdirs dyn_classdirs static_classdirs
	
	for class in $recipe_classes default utils; do
		if [[ " $all_classes " =~ " $class " ]]; then continue; fi
		all_classes="$all_classes $class"
		if [[ " $all_dynamic_classes " =~ " $class " ]]; then 
			dyn_classes="$dyn_classes $class"
		else
			static_classes="$static_classes $class"
		fi
		classdir="$(script_class_dir $class)"
		if [[ " $all_classdirs " =~ " $classdir " ]]; then continue; fi
		all_classdirs="$all_classdirs $classdir"
		if [[ " $all_dynamic_classes " =~ " $class " ]]; then 
			dyn_classdirs="$dyn_classdirs $classdir"
		else
			static_classdirs="$static_classdirs $classdir"
		fi
	done

	# First look for a recipe-specific script
	dir="$recipe_metadatadir/buildsteps/scripts"
	if [ -d "$dir" ]; then
		echo "$dir"
	fi

	# Then look for a class-specific scripts
	dir="$PLATFORM_LOC/classes"
	# Search for files in the platform "classes" directory
	if [ -d "$dir" ]; then
		# Search for files in a class specific subdirectory (here use the class name, even for
		# static classes, that are anyway searched in a $class subdirectory of their plugin)
		echo "$dir"
		for classdir in $all_classes; do
			if [ -d "$dir/$classdir" ]; then
				echo "$dir/$classdir"
			fi
		done
	fi
	# Then in the platform "scripts/plugins" directory
	dir="$PLATFORM_LOC/scripts/plugins"
	if [ -d "$dir" ]; then
		# First static plugin/class specific subdirectory
		for class in $static_classes; do
			classdir=$(script_class_dir $class)
			if [ -d "$dir/$classdir/classes/$class" ]; then
				echo "$dir/$classdir/classes/$class"
			fi
		done
		# Then dynamic class subdirectories
		for class in $dyn_classes; do
			if [ -d "$dir/$dynamic_classes_plugin/$class" ]; then
				echo "$dir/$dynamic_classes_plugin/$class"
			fi
		done
	fi
	# And in installed "classes" directories
	if [ -d "$sw4l_install_base" ]; then
		for class in $static_classes; do
			classdir=$(script_class_dir $class)
			subdir=
			# Search for the most recent plugin directory
			for dir in "$sw4l_install_base/plugins/fr.ac6.${classdir}_"*; do
				if [ -d "$dir" ]; then
					subdir="$dir"
				fi
			done
			if [ -n "$subdir" -a -d "$subdir/resources/scripts/classes/$class" ]; then
				echo "$subdir/resources/scripts/classes/$class"
			fi
			
			# In debug mode scripts location is different
			if [ "$DEBUG_SW4L" == true ]; then
				for dir in "$sw4l_install_base"; do
					if [ -d "$dir" ]; then
						subdir="$dir"
					fi
				done
				if [ -n "$subdir" -a -d "$subdir/classes/$class" ]; then
					echo "$subdir/classes/$class"
				fi
			fi
		done
		if [ -d "$dynamic_classes_plugin_dir" ]; then
			if [ -n "$dynamic_classes_plugin_subdir" -a "$dynamic_classes_plugin_subdir" != "." ]; then
				dir="$dynamic_classes_plugin_dir/$dynamic_classes_plugin_subdir"
			else
				dir="$dynamic_classes_plugin_dir"
			fi
			for classdir in $dyn_classdirs; do
				if [ -d "$dir/$classdir" ]; then
					echo "$dir/$classdir"
				fi
			done
		fi
	fi

	# Then search for generic scripts in the platform "scripts" directory
	dir="$PLATFORM_LOC/scripts"
	if [ -d "$dir" ]; then
		# First for files directly in "scripts"
		echo "$dir"
		# then for files in a static plugin/class specific subdirectory
		for classdir in $static_classdirs; do
			if [ -d "$dir/plugins/$classdir" ]; then
				# first in classes/class-name
				if [ -d "$dir/plugins/$classdir/classes/$class" ]; then
					echo "$dir/plugins$classdir//classes/$class"
				fi
				# then directly in the plugin scripts directory
				echo "$dir/plugins/$classdir"
			fi
		done
	fi
	
	# And finally in the installed plugins
	if [ -d "$sw4l_install_base" ]; then
		# Note that originaly only main scripts were located in class-specific plugins
		# and searched directly by SW4L Java code (that must be updated to look not only
		# in $PLATFORM_LOC/scripts, but also in the plugins subdirectories)
		# Most included files are in linux.rootfs or linux.utils, included last
		# but dynamic-class-specific directories should be searched first
		if [ -d "$dynamic_classes_plugin_dir" ]; then
			if [ -n "$dynamic_classes_plugin_subdir" -a "$dynamic_classes_plugin_subdir" != "." ]; then
				dir="$dynamic_classes_plugin_dir/$dynamic_classes_plugin_subdir"
			else
				dir="$dynamic_classes_plugin_dir"
			fi
			for classdir in $dyn_classdirs; do
				if [ -d "$dir/$classdir" ]; then
					echo "$dir/$classdir"
				fi
			done
		fi
		for class in $static_classes; do
			classdir=$(script_class_dir $class)
			subdir=
			# Search for the most recent plugin directory
			for dir in "$sw4l_install_base/plugins/fr.ac6.${classdir}_"*; do
				if [ -d "$dir" ]; then
					subdir="$dir"
				fi
			done
			if [ -n "$subdir" -a -d "$subdir/resources/scripts" ]; then
				if [ "$subdir/resources/scripts" = "$lastchance" ]; then lastchance=""; fi
				echo "$subdir/resources/scripts"
			fi
		done
	fi
	if [ -n "$lastchance" ]; then
		echo "$lastchance"
	fi
}
scripts_search_path=$(setup_scripts_dir)

#
# Search for scripts
# 	If arguments contain spaces, they are split and name several scripts to search for
#	Script names are automatically appended a '.sh' extension if needed
#	Underscores and dashes are equivalent in script names, but underscores take precedence
#
search_all_scripts() {
	local scripts=" $* "
	dbg scripts "Searching for all scripts from '$scripts'" 
	scripts="${scripts//.sh / }"
	scripts="${scripts// do_/ }"
	scripts="${scripts//[-_]/_} ${scripts//[-_]/-} ${scripts}"
	local scriptdir 
	for scriptdir in $scripts_search_path; do
		local script
		for script in $scripts; do
			local file
			for file in $scriptdir/$script \
			            $scriptdir/$script.sh  \
			            $scriptdir/do_$script  \
			            $scriptdir/do-$script.sh
			do
				#dbg scripts "Looking for script '$file'" 
				if [ -r "$file" ]; then
					echo "$file"
				fi
			done
		done
	done
}

search_all_first_scripts() {
	local scripts=" $* "
	dbg scripts "Searching for all first scripts from '$scripts'" 
	scripts="${scripts//.sh / }"
	scripts="${scripts// do_/ }"
	scripts="${scripts//[-_]/_} ${scripts//[-_]/-} ${scripts}"
	local scriptdir 
	for scriptdir in $scripts_search_path; do
		local script
		for script in $scripts; do
			local file
			for file in $scriptdir/$script \
			            $scriptdir/$script.sh  \
			            $scriptdir/do_$script  \
			            $scriptdir/do-$script.sh
			do
				#dbg scripts "Looking for script '$file'" 
				if [ -r "$file" ]; then
					echo "$file"
					break
				fi
			done
		done
	done
}

search_script() {
	local scripts=" $* "
	dbg scripts "Searching for script '$scripts'" 
	scripts="${scripts//.sh / }"
	scripts="${scripts// do_/ }"
	scripts="${scripts//[-_]/_} ${scripts//[-_]/-} ${scripts}"
	local scriptdir 
	for scriptdir in $scripts_search_path; do
		local script
		for script in $scripts; do
			local file
			for file in $scriptdir/$script \
			            $scriptdir/$script.sh  \
			            $scriptdir/do_$script  \
			            $scriptdir/do-$script.sh
			do
				#dbg scripts "Looking for script '$file'" 
				if [ -r "$file" ]; then
					echo "$file"
					return
				fi
			done
		done
	done
}

do_require_script() {
	local scripts="$*"
	local found=$(search_script "$@")
	if [ -z "$found" ]; then
		local message="Cannot find any of scripts"
		declare -a script
		read -a script <<< "$scripts"
		local sep i
		for ((i=0; i < ${#script[@]} - 1; i++)) do
			message="$message$sep '${script[i]}'"
			sep=","
		done
		if [ -n "$sep" ]; then sep=" or"; fi
		message="$message$sep '${script[i]}'"
		# Can't use error here: this function is needed to load log_utils.sh...
		fatal 1 "$message"
	fi
	echo "$found"
}

load_script_once() {
	local script="$1"
	if record_script_load "$script"; then
		dbg_note "Loading script '$script'"
		source "$script"
		local status=$?
		dbg scripts "Finished loading script '$script' - status=$status"
		return $status
	fi
	return 0 # Script was already loaded
}

load_script() {
	local script="$1"
	record_script_load "$script"
	dbg_note "Loading script '$script'"
	source "$script"
	local status=$?
	dbg scripts "Finished loading script '$script' - status=$status"
	return $status
}

load_script_log() {
	local script="$1"
	record_script_load "$script"
	note "Loading script '$script'"
	source "$script"
	local status=$?
	dbg_note scripts "Finished loading script '$script' - status=$status"
	return $status
}

require_script_once() {
	local found=$(do_require_script "$@")
	load_script_once "$found"
}

require_script() {
	local found=$(do_require_script "$@")
	load_script "$found"
}

source_script_once() {
	local found=$(search_script "$@")
	if [ -z "$found" ]; then
		return 0 # Not finding the script is NOT an error, it's expected
	fi
	load_script_once "$found"
}

source_script() {
	local found=$(search_script "$@")
	if [ -z "$found" ]; then
		return 0 # Not finding the script is NOT an error, it's expected
	fi
	load_script "$found"
}

source_script_log() {
	local found=$(search_script "$@")
	if [ -z "$found" ]; then
		return 0 # Not finding the script is NOT an error, it's expected
	fi
	load_script_log "$found"
}

source_all_scripts() {
	local script
	for script in $(search_all_scripts "$@"); do
		note "Loading script '$script'"
		load_script_log "$script"
		local status=$?
		dbg_note "Done loading script '$script': status=$status"
		if [ $status -ne 0 ]; then return $status; fi
	done
}

source_all_first_scripts() {
	local script
	for script in $(search_all_first_scripts "$@"); do
		note "Loading script '$script'"
		load_script_log "$script"
		local status=$?
		dbg_note "Done loading script '$script': status=$status"
		if [ $status -ne 0 ]; then return $status; fi
	done
}

source_all_scripts_reverse() {
	local script
	for script in $(search_all_scripts "$@" | tac); do
		note "Loading script '$script'"
		load_script_log "$script"
		local status=$?
		dbg_note "Done loading script '$script': status=$status"
		if [ $status -ne 0 ]; then return $status; fi
	done
}

source_all_first_scripts_reverse() {
	local script
	for script in $(search_all_first_scripts "$@" | tac); do
		note "Loading script '$script'"
		load_script_log "$script"
		local status=$?
		dbg_note "Done loading script '$script': status=$status"
		if [ $status -ne 0 ]; then return $status; fi
	done
}

call_buildstep_prepend() {
	local scripts="$1"
	local pfx="$2"
	local script prepend after fprepend fafter
	for script in $scripts; do
		prepend="$prepend ${script}-prepend"
		fprepend="$fprepend ${pfx}${script}-prepend"
		fafter="$fafter ${pfx}${script}-prepend_after"
	done

	call_all_existing_reverse "$fprepend"
	status=$?
	if [ $status -ne 0 ]; then return $status; fi
	source_script_log "$prepend"
	status=$?
	if [ $status -ne 0 ]; then return $status; fi
	call_all_existing "$fafter"
}

do_buildstep_prepend() {
	local status=$? class
	if [ $status -ne 0 ]; then return $status; fi
	if [ "$buildstep_script" != "$current_buildstep" ]; then
		call_buildstep_prepend "${buildstep_script} ${current_buildstep}" "do_"
	else
		call_buildstep_prepend "${current_buildstep}" "do_"
	fi
}

call_buildsubstep()  {
	local status=$? 
	if [ $status -ne 0 ]; then return $status; fi
	
	local buildsubstep="$1"
	shift
	call_buildstep_prepend "$buildsubstep"
	status=$?
	if [ $status -ne 0 ]; then return $status; fi
	call_first_existing "$buildsubstep" "$@"
	status=$?
	if [ $status -ne 0 ]; then return $status; fi
	call_buildstep_append "$buildsubstep"
}

do_buildstep()  {
	local status=$? class
	if [ $status -ne 0 ]; then return $status; fi
	if [ "${buildstep_script#do_}" != "$current_buildstep" ]; then
		buildstep="do_${buildstep_script#do_} do_${current_buildstep}"
	else
		buildstep="do_${current_buildstep}"
	fi
	shell_function_executed=false
	call_first_existing "$buildstep" "$@"
	status=$?
	if [ $status -ne 0 ]; then return $status; fi
	if ! $shell_function_executed; then
		log "Nothing to do for buildstep '${current_buildstep}' of '$recipe_name'"
	fi
	return 0
}

call_buildstep_append() {
	local scripts="$1"
	local pfx="$2"
	local script append before fappend fbefore
	for script in $scripts; do
		append="$append ${script}-append"
		fappend="$fappend ${pfx}${script}-append"
		fbefore="$fbefore ${pfx}${script}-append_before"
	done

	call_all_existing_reverse "$fbefore"
	status=$?
	if [ $status -ne 0 ]; then return $status; fi
	source_script_log "$append"
	status=$?
	if [ $status -ne 0 ]; then return $status; fi
	call_all_existing "$fappend"
}

do_buildstep_append() {
	local status=$? class
	if [ $status -ne 0 ]; then return $status; fi
	if [ "$buildstep_script" != "$current_buildstep" ]; then
		call_buildstep_append "${buildstep_script} ${current_buildstep}" "do_"
	else
		call_buildstep_append "${current_buildstep}" "do_"
	fi
}

######################################################################################################################
#                                                                                                                    #
#                                          Source specific script routines                                           #
#                                                                                                                    #
######################################################################################################################

require_script_once package_utils.sh # Needed for package name utilities
require_script_once path_utils.sh
require_script_once task_utils.sh

# Also include class-specific utilities, if any, first the more generic, then the more specific
for class in $recipe_classes_reversed; do
	source_script_once ${class}_utils.sh
done

# Finally also include buildstep-specific utilities, if any
source_script_once ${current_buildstep}_utils.sh 
if [ "${buildstep_script}" != "${current_buildstep}" ]; then
	source_script_once ${buildstep_script}_utils.sh 
fi

#	
# To take care of possible buildstep locking requirements, like
# for example for fetch and import, if a variable named
# BUILDSTEP_LOCK_FILE exist, then the buildstep is executed with
# the given file locked; if the lock spec starts by a + sign
# the lock is shared, otherwise exclusive
#
if [ -n "$BUILDSTEP_LOCK_FILE" ]; then
	exec 9>> "${BUILDSTEP_LOCK_FILE#+}"
	if [ "$BUILDSTEP_LOCK_FILE" != "${BUILDSTEP_LOCK_FILE#+}" ]; then
		note "$(date -Ins): Placing shared lock on ${BUILDSTEP_LOCK_FILE#+}"
		flock -s 9
	else
		note "$(date -Ins): Placing exclusive lock on ${BUILDSTEP_LOCK_FILE#+}"
		flock -x 9
	fi
	note "$(date -Ins): Lock acquired"
fi

######################################################################################################################
#                                                                                                                    #
#                                           End of generic script routines                                           #
#                                                                                                                    #
######################################################################################################################
note "Executing in $(/bin/pwd)"
