######################################################################################################################
#                                                                                                                    #
#                                                     Task utilities                                                 #
#                                                                                                                    #
######################################################################################################################

# Global configuration variables
: ${possible_terminals:=gnome-terminal konsole rxvt xfce xterm}
: ${default_terminal:=auto}
: ${recipe_statedir:=.ac6_build}

# Set the current status to some saved value
set_status() { 
	local status=${1:=0}
	return $status
}

#
# recipe_task_state_file TASK RECIPE
# 	Get the state file base name for task TASK (defaults to current task)
#	of recipe RECIPE (defaults to current recipe)
#
recipe_task_state_file() {
	local task="$1"
	local recipe="$2"
	[ -n "$task" ] || task="$current_task"
	[ -n "$recipe" ] || recipe="$recipe_name"
	task="${task,,}"
	task="${task/ /-}"
	if [ "$recipe" = "$recipe_name" -a "$task" = "$current_task" ]; then
		mkdir -p "${taskBuildStatusDir}/${recipe_statedir}"
		echo "${taskBuildStatusDir}/${recipe_statedir}/task.${task}.${recipe_id}"
		return 0
	fi
	local recipe_dir
	local recipe_spec
	local dir
	for dir in ../$recipe*; do
		if [ -d $dir ]; then
			recipe_dir=$dir/_build
			recipe_spec=${dir##*/}
			break
		fi
	done
	if [ -z "$recipe_spec" ]; then
		recipe_spec=${recipe} # Only possible if task is fetch or import
		recipe_dir=$tmpdir/_build # Fake an unexistant dir
	fi
	# This mimics the taskBuildStatusDir definition by the IDE
	case $task in
		populate-sysroot)
			mkdir -p "${SYSROOT_LOC}/${recipe_statedir}"
			echo "${SYSROOT_LOC}/${recipe_statedir}/task.${task}.${recipe_spec}"
			;;
		split | package)
			mkdir -p "${PLATFORM_LOC}/feeds/${recipe_statedir}"
			echo "${PLATFORM_LOC}/feeds/${recipe_statedir}/task.${task}.${recipe_spec}"
			;;
		fetch)
			mkdir -p "${DOWNLOAD_LOC}/${recipe_statedir}"
			echo "${DOWNLOAD_LOC}/${recipe_statedir}/task.${task}.${recipe_spec}"
			;;
		*)
			mkdir -p "${recipe_dir}/${recipe_statedir}"
			echo "${recipe_dir}/${recipe_statedir}/task.${task}.${recipe_spec}"
			;;
	esac
}

#
# poison_recipe_task TASK RECIPE
# 	Force task TASK (defaults to current task) of recipe
#	RECIPE (defaults to current recipe) to be re-executed
#
poison_recipe_task() {
	local task_state_file=$(recipe_task_state_file "$@")
	# Only poison a task that was already executed...
	if [ -s "$task_state_file" ]; then
		task_execute_date > $task_state_file.poison
	fi
}

#
# generate_task_output_state
#	Generate current task output state, overriding the default state stamp
#
generate_task_output_state() {
	if [  $# -eq 0 ]; then
		internal_error "generate_task_output_state needs some value to put in the output state file"
	fi
	local task_state_file=$(recipe_task_state_file)
	echo "$*" > $task_state_file.outstamp
}

#
# generate_other_task_output_state
#	Generate output state for another task, overriding the default state stamp
#
generate_other_task_output_state() {
	local task="$1"
	shift
	if [  $# -eq 0 ]; then
		internal_error "generate_other_task_output_state needs some value to put in the output state file"
	fi
	local task_state_file=$(recipe_task_state_file $task)
	echo "$*" > $task_state_file.outstamp
}

#
# remove_task_output_state TASK RECIPE
#	Remove task TASK (defaults to current task) of recipe RECIPE 
#	(defaults to current recipe) output state, usually only needed if
#	there is a risk dependent tasks may be marked for systematic reexecution
remove_task_output_state() {
	local task_state_file=$(recipe_task_state_file "$@")
	rm -f $task_state_file.outstamp
}

#
# poison_dependent_tasks
#	Force tasks depending tasks on the current task to be systematically re-executed
#
# Called by a task to force execution of all its dependents, usually another
# task of the same recipe, which itself will generate an output state value
#
poison_dependent_tasks() { 
	local task_state_file=$(recipe_task_state_file)
	rm -f $task_state_file.outstamp
	touch $task_state_file.outstamp
}

# Currently doe not work?
unconditional_task() {
	local task_state_file=$(recipe_task_state_file)
	# Poison myself even if I was not yet executed...
	mkdir -p $(dirname $task_state_file.poison)
	rm -f $task_state_file.poison
	touch $task_state_file.poison
}

first_executable() {
	local exe file
	for exe; do
		file="$(which $exe)"
		if [ -n "$file" -a -x "$file" ]; then
			echo "$exe"
			break
		fi
	done
}

do_launch_in_background() {
	local cmd="$*"
	trap "exit 0" SIGTERM
	(
		:
		local pid=$BASHPID
		bash -c "sleep 2; kill -TERM $pid" >/dev/null 2>&1 &
		local sleepid=$!
		# redirects this subshell output to /dev/null so that Eclipse will not wait for children termination
		exec bash -c "$cmd; status=\$?; kill -9 $sleepid; exit \$status" >/dev/null 2>&1
	)  >/dev/null 2>&1
	local status=$?
	if [ $status -eq 143 ]; then # 143 == SIGTERM signal (128 + 15)
		return 0
	else
		return $status
	fi
}

launch_in_background() { 
	do_launch_in_background "$@" > /dev/null 2>&1
}

execute_buildstep() {
	local scriptdir script scripts script_file name="${1%.sh}"
	shift
	case $name in
		*.sh)				script="${name%.sh}"	;;
		development-shell)	script="shell"			;;
		compile)			script="make_all"		;;
		install)			script="make_install"	;;
		apply-patches)		script="patch"			;;
		*)					script="${name//-/_}"	;;
	esac
	for class in $recipe_classes; do
		scripts="$scripts $class-$script.sh"
	done
	scripts="$scripts $script.sh do_$script.sh"
	for file in $scripts; do
		for scriptdir in $scripts_search_path; do
			if [ -r "$scriptdir/$file" ]; then
				script_file="$scriptdir/$file"
				break
			fi
		done
		if [ -n "$script_file" ]; then break; fi
	done
	if [ -z "$script_file" ]; then
		echo "ERROR: Can't find script for buildstep '$name'"
		return 127
	fi
	local buildstep=$(get_current_buildstep $script)
	: ${buildstep:=$name}
	echo BUILDSTEP_SCRIPT_PATH=$script_file CURRENT_BUILDSTEP=$buildstep bash $script_file "$@"
	BUILDSTEP_SCRIPT_PATH=$script_file CURRENT_BUILDSTEP=$buildstep bash $script_file "$@"
}

try_launch_terminal() {
	local exe="$1"
	shift
	local bash_cmd="$*"
	file="$(which $exe)"
	if [ -n "$file" -a -x "$file" ]; then
		export -f execute_buildstep
		export -f get_current_buildstep
		export scripts_search_path
		case $exe in
			gnome-terminal)
				# gnome terminal automatically forks in background, so wait termination
				# using a FIFO; note that if using --wait, gnome-terminal exits ONLY if
				# the forked command terminates normally, NOT if you close the window, so
				# we can't use --wait here...
				
				# Create a pipe I will read to wait for son to finish
				local wait_fifo_name="/tmp/wait.$$"
				mkfifo $wait_fifo_name
				
				# Now starts the gnome terminal
				$exe -- bash -c "exec $bash_cmd 9> $wait_fifo_name"

				local status=$?
				if [ $status -ne 0 ]; then
					rm -f $wait_fifo_name
					return $status
				fi

				sleep 1 # Ensures the FIFO is opened for writing
				cat < $wait_fifo_name
				rm -f $wait_fifo_name
				;;
			konsole)
				eval $exe --nofork -e $bash_cmd
				;;
			*)
				eval $exe -e $bash_cmd
				;;
		esac
	fi
}
	
launch_terminal() {
	local exe file terminals=$possible_terminals wrapper
	local motd_file="$1" bash_cmd="bash"
	if [ -n "${default_terminal}" -a "${default_terminal}" != "auto" ]; then
		terminals=$default_terminal
	fi
	if [ -n "$motd_file" -a -r "$motd_file" ]; then
		wrapper=.bashrc-wrapper.$$
		cat <<EOF > $wrapper
#
# Wrapper to print a oneshot welcome message when starting bash
#

if [ -r "$motd_file" ]; then
	cat "$motd_file"
	rm -f "$motd_file"
fi

if [ -r ~/.bashrc ]; then
	source ~/.bashrc
fi
rm -f \$BASH_SOURCE
EOF
		bash_cmd="bash --rcfile $wrapper"
	fi
	for exe in $possible_terminals; do
		if try_launch_terminal $exe $bash_cmd; then
			return 0
		fi
	done
	error 127 "Don't find any working terminal in '$possible_terminals'"
	return 1
}

try_execute_in_terminal() {
	local term=$1
	shift
	local cmd_opt="-e"
	local other_opt=""
	local quote_cmd=false
	local wait=false
	local wait_fifo_name wait_fifo
	case $term in
		gnome-terminal)
			# quote_cmd=true
			wait=true
			cmd_opt="-x"
			;;
		konsole)
			other_opt="--nofork"
			;;
	esac
	if $wait; then
		# Create a pipe I will read to wait for son to finish
		wait_fifo_name="/tmp/wait.$$"
		mkfifo $wait_fifo_name
		wait_fifo="2> $wait_fifo_name"
	fi
	if $quote_cmd; then
		$term $other_opt $cmd_opt "$* $wait_fifo"
	else
		$term $other_opt $cmd_opt /bin/bash -c "$* $wait_fifo"
	fi
	local status=$?
	if [ $status -ne 0 ]; then
		rm -f $wait_fifo_name
		return $status
	fi
	if $wait; then
		sleep 1 # Ensures the FIFO is opened for writing
		cat < $wait_fifo_name
		rm -f $wait_fifo_name
	fi
}

execute_in_terminal() { 
	local exe file terminals=$possible_terminals
	if [ -n "${default_terminal}" -a "${default_terminal}" != "auto" ]; then
		terminals=$default_terminal
	fi
	for exe in $possible_terminals; do
		file="$(which $exe)"
		if [ -n "$file" -a -x "$file" ]; then
			echo try_execute_in_terminal $exe "$@"
			if try_execute_in_terminal $exe "$@"; then return 0; fi
		fi
	done
	error "Don't find any working terminal in '$possible_terminals'"
	return 1
}
