######################################################################################################################
#                                                                                                                    #
#                                             User group add utilities                                            	 #
#										  (Require shadow-native to be built)										 #
#								The following functions should be run under fakeroot						 		 #
#                                                                                                                    #
######################################################################################################################

###########################################################################
#
# Create a new user in to rootfs
#
#	perform_useradd
#		rootdir			output rootfs root location
#		opts			useradd parameters
#
###########################################################################
perform_useradd() {
	local rootdir="$1"
	local opts="$2"
	
	log "${PN}: Performing useradd with [$opts]"
	local username=`echo "$opts" | awk '{ print $NF }'`
	local user_exists="`grep "^$username:" $rootdir/etc/passwd || true`"
	
	if test "x$user_exists" = "x"; then
		eval flock -x $rootdir${sysconfdir} -c  \"useradd \$opts\" || true
		user_exists="`grep "^$username:" $rootdir/etc/passwd || true`"
		if test "x$user_exists" = "x"; then
			fatal "${PN}: useradd command did not succeed."
		fi
	else
		log "${PN}: user $username already exists, not re-creating it"
	fi
}

###########################################################################
#
# Create new group in to rootfs
#
#	perform_useradd
#		rootdir			output rootfs root location
#		opts			groupadd parameters
#
###########################################################################
perform_groupadd () {
	local rootdir="$1"
	local opts="$2"
	
	log "${PN}: Performing groupadd with [$opts]"
	local groupname=`echo "$opts" | awk '{ print $NF }'`
	local group_exists="`grep "^$groupname:" $rootdir/etc/group || true`"
	
	if test "x$group_exists" = "x"; then
		eval flock -x $rootdir${sysconfdir} -c \"groupadd \$opts\" || true
		group_exists="`grep "^$groupname:" $rootdir/etc/group || true`"
		if test "x$group_exists" = "x"; then
			fatal "${PN}: groupadd command did not succeed."
		fi
	else
		log "${PN}: group $groupname already exists, not re-creating it"
	fi
}

###########################################################################
#
# Set group membership in to rootfs
#
#	perform_useradd
#		rootdir			output rootfs root location
#		opts			groupmems parameters
#
###########################################################################
perform_groupmems () {
	local rootdir="$1"
	local opts="$2"
	log "${PN}: Performing groupmems with [$opts]"
	local groupname=`echo "$opts" | awk '{ for (i = 1; i < NF; i++) if ($i == "-g" || $i == "--group") print $(i+1) }'`
	local username=`echo "$opts" | awk '{ for (i = 1; i < NF; i++) if ($i == "-a" || $i == "--add") print $(i+1) }'`
	log "${PN}: Running groupmems command with group $groupname and user $username"
	local mem_exists="`grep "^$groupname:[^:]*:[^:]*:\([^,]*,\)*$username\(,[^,]*\)*$" $rootdir/etc/group || true`"
	if test "x$mem_exists" = "x"; then
		eval flock -x $rootdir${sysconfdir} -c \"groupmems \$opts\" || true
		mem_exists="`grep "^$groupname:[^:]*:[^:]*:\([^,]*,\)*$username\(,[^,]*\)*$" $rootdir/etc/group || true`"
		if test "x$mem_exists" = "x"; then
			fatal "${PN}: groupmems command did not succeed."
		fi
	else
		log "${PN}: group $groupname already contains $username, not re-adding it"
	fi
}

###########################################################################
#
# Delete group from rootfs
#
#	perform_useradd
#		rootdir			output rootfs root location
#		opts			groupdel parameters
#
###########################################################################
perform_groupdel () {
	local rootdir="$1"
	local opts="$2"
	log "${PN}: Performing groupdel with [$opts]"
	local groupname=`echo "$opts" | awk '{ print $NF }'`
	local group_exists="`grep "^$groupname:" $rootdir/etc/group || true`"

	if test "x$group_exists" != "x"; then
		local awk_input='BEGIN {FS=":"}; $1=="'$groupname'" { print $3 }'
		local groupid=`echo "$awk_input" | awk -f- $rootdir/etc/group`
		local awk_check_users='BEGIN {FS=":"}; $4=="'$groupid'" {print $1}'
		local other_users=`echo "$awk_check_users" | awk -f- $rootdir/etc/passwd`

		if test "x$other_users" = "x"; then
			eval flock -x $rootdir${sysconfdir} -c \"groupdel \$opts\" || true
			group_exists="`grep "^$groupname:" $rootdir/etc/group || true`"
			if test "x$group_exists" != "x"; then
				fatal "${PN}: groupdel command did not succeed."
			fi
		else
			log "${PN}: '$groupname' is primary group for users '$other_users', not removing it"
		fi
	else
		log "${PN}: group $groupname doesn't exist, not removing it"
	fi
}

###########################################################################
#
# Delete user from rootfs
#
#	perform_useradd
#		rootdir			output rootfs root location
#		opts			userdel parameters
#
###########################################################################
perform_userdel () {
	local rootdir="$1"
	local opts="$2"
	log "${PN}: Performing userdel with [$opts]"
	local username=`echo "$opts" | awk '{ print $NF }'`
	local user_exists="`grep "^$username:" $rootdir/etc/passwd || true`"
	if test "x$user_exists" != "x"; then
		eval flock -x $rootdir${sysconfdir} -c \"userdel \$opts\" || true
		user_exists="`grep "^$username:" $rootdir/etc/passwd || true`"
		if test "x$user_exists" != "x"; then
			fatal "${PN}: userdel command did not succeed."
		fi
	else
		log "${PN}: user $username doesn't exist, not removing it"
	fi
}

###########################################################################
#
# Modify a group definition on the rootfs
#
# This script must define one function, called with the following parameters
#
#	perform_useradd
#		rootdir			output rootfs root location
#		opts			groupmod parameters
#
###########################################################################
perform_groupmod () {
	# Other than the return value of groupmod, there's no simple way to judge whether the command
	# succeeds, so we disable -e option temporarily
	set +e
	local rootdir="$1"
	local opts="$2"
	log "${PN}: Performing groupmod with [$opts]"
	local groupname=`echo "$opts" | awk '{ print $NF }'`
	local group_exists="`grep "^$groupname:" $rootdir/etc/group || true`"
	if test "x$group_exists" != "x"; then
		eval flock -x $rootdir${sysconfdir} -c \"$groupmod \$opts\"
		if test $? != 0; then
			warning "${PN}: groupmod command did not succeed."
		fi
	else
		warning "${PN}: group $groupname doesn't exist, unable to modify it"
	fi
	set -e
}

###########################################################################
#
# Modify a user definition on the rootfs
#
# This script must define one function, called with the following parameters
#
#	perform_useradd
#		rootdir			output rootfs root location
#		opts			usermod parameters
#
###########################################################################
perform_usermod () {
	# Same reason with groupmod, temporarily disable -e option
	set +e
	local rootdir="$1"
	local opts="$2"
	log "${PN}: Performing usermod with [$opts]"
	local username=`echo "$opts" | awk '{ print $NF }'`
	local user_exists="`grep "^$username:" $rootdir/etc/passwd || true`"
	if test "x$user_exists" != "x"; then
		eval flock -x $rootdir${sysconfdir} -c \"$usermod \$opts\"
		if test $? != 0; then
			fatal "${PN}: usermod command did not succeed."
		fi
	else
		warning "${PN}: user $username doesn't exist, unable to modify it"
	fi
	set -e
}

useradd_preinst () {
	SYSROOT="/"
	OPT=""
	
	# If we're not doing a special SSTATE/SYSROOT install
	# then set the values, otherwise use the environment
	if test "x$UA_SYSROOT" = "x"; then
		SYSROOT="$ROOTFS_LOC"
		OPT="--root $ROOTFS_LOC"
		
		# Installing onto a target
		# Add groups and users defined only for this package
		GROUPADD_PARAM="%GROUPADD_PARAM%"
		USERADD_PARAM="%USERADD_PARAM%"
		GROUPMEMS_PARAM="%GROUPMEMS_PARAM%"
	fi
	
	# Perform group additions first, since user additions may depend
	# on these groups existing
	if test "x`echo $GROUPADD_PARAM | tr -d '[:space:]'`" != "x"; then
		echo "Running groupadd commands..."
		# Invoke multiple instances of groupadd for parameter lists
		# separated by ';'
		opts=`echo "$GROUPADD_PARAM" | cut -d ';' -f 1 | sed -e 's#[ \t]*$##'`
		remaining=`echo "$GROUPADD_PARAM" | cut -d ';' -f 2- | sed -e 's#[ \t]*$##'`
		while test "x$opts" != "x"; do
			perform_groupadd "$SYSROOT" "$OPT $opts"
			if test "x$opts" = "x$remaining"; then
				break
			fi
			opts=`echo "$remaining" | cut -d ';' -f 1 | sed -e 's#[ \t]*$##'`
			remaining=`echo "$remaining" | cut -d ';' -f 2- | sed -e 's#[ \t]*$##'`
		done
	fi 
	
	if test "x`echo $USERADD_PARAM | tr -d '[:space:]'`" != "x"; then
		echo "Running useradd commands..."
		# Invoke multiple instances of useradd for parameter lists
		# separated by ';'
		opts=`echo "$USERADD_PARAM" | cut -d ';' -f 1 | sed -e 's#[ \t]*$##'`
		remaining=`echo "$USERADD_PARAM" | cut -d ';' -f 2- | sed -e 's#[ \t]*$##'`
		while test "x$opts" != "x"; do
			perform_useradd "$SYSROOT" "$OPT $opts"
			if test "x$opts" = "x$remaining"; then
				break
			fi
			opts=`echo "$remaining" | cut -d ';' -f 1 | sed -e 's#[ \t]*$##'`
			remaining=`echo "$remaining" | cut -d ';' -f 2- | sed -e 's#[ \t]*$##'`
		done
	fi
	
	if test "x`echo $GROUPMEMS_PARAM | tr -d '[:space:]'`" != "x"; then
		echo "Running groupmems commands..."
		# Invoke multiple instances of groupmems for parameter lists
		# separated by ';'
		opts=`echo "$GROUPMEMS_PARAM" | cut -d ';' -f 1 | sed -e 's#[ \t]*$##'`
		remaining=`echo "$GROUPMEMS_PARAM" | cut -d ';' -f 2- | sed -e 's#[ \t]*$##'`
		while test "x$opts" != "x"; do
			perform_groupmems "$SYSROOT" "$OPT $opts"
			if test "x$opts" = "x$remaining"; then
				break
			fi
			opts=`echo "$remaining" | cut -d ';' -f 1 | sed -e 's#[ \t]*$##'`
			remaining=`echo "$remaining" | cut -d ';' -f 2- | sed -e 's#[ \t]*$##'`
		done
	fi
}

do_useradd_rootfs () {
	local pkgId=$1

	# It is also possible we may be in a recipe which doesn't have useradd dependencies and hence the
	# useradd/groupadd tools are unavailable. If there is no dependency, we assume we don't want to
	# create users in the sysroot
	if ! command -v useradd; then
		warning "command useradd not found!"
		exit 0
	fi

	# Add groups and users defined for all recipe packages
	GROUPADD_PARAM="${@get_all_cmd_params(d, 'groupadd')}"
	USERADD_PARAM="${@get_all_cmd_params(d, 'useradd')}"
	GROUPMEMS_PARAM="${@get_all_cmd_params(d, 'groupmems')}"

	# Tell the system to use the environment vars
	UA_SYSROOT=1

	useradd_preinst
}
